(   Title:  A utiilty for testing words that use the frstack for
	    MPFR multiple precision float-reliables
     File:  frtester.fs
   Author:  David N. Williams
  License:  Public Domain, John Hayes
  Version:  1.5.0
  Started:  March 26, 2011
  Revised:  March 26, 2011

For any code derived from John Hayes' tester program:
)
\ (C) 1995 JOHNS HOPKINS UNIVERSITY / APPLIED PHYSICS LABORATORY
\ MAY BE DISTRIBUTED FREELY AS LONG AS THIS COPYRIGHT NOTICE REMAINS.
(
This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

The rest is public domain.
)
BASE @ DECIMAL

\ *** ERRORS

VARIABLE FRT-ERROR-XT
: FRT-ERROR FRT-ERROR-XT @ EXECUTE ;  \ vectored error reporting

VARIABLE frt-error-index  \ holds array index|-1 after error

: FRT-ERROR-DEFAULT   ( c-addr u -- )
(
Display an error message followed by the line that had the
error.
)
  TYPE SOURCE TYPE CR ;
' FRT-ERROR-DEFAULT FRT-ERROR-XT !

\ *** COMPARISONS

VARIABLE FRT-TEST=-XT

: FRT-TEST=  (fr: x y -- ) ( -- flag )  FRT-TEST=-XT @ EXECUTE ;

\ The sign of these tolerances must be plus.
FRVARIABLE FRT-ABS-ERROR   fr" 0e0"   FRT-ABS-ERROR  fr!
FRVARIABLE FRT-REL-ERROR   fr" 1e-12" FRT-REL-ERROR  fr!

: FRT-DATUM=  (fr: x y -- ) ( -- flag )
(
Leave TRUE if the two frfloats have equivalent internal
representations.  Else leave FALSE.
)
  fr= ;

\ |x - y| < eps
: FRT-ABS=  (fr: x y -- ) ( -- flag )
  fr- |fr| FRT-ABS-ERROR fr@ fr<  ;

\ |m - r| < eps * |r|
: FRT-REL=  (fr: meas ref -- ) ( -- flag )
  frrel- FRT-REL-ERROR fr@ fr< ;

: FRT-REL0=  (fr: meas ref -- ) ( -- flag )
  frdup fr0= IF FRT-ABS= ELSE FRT-REL= THEN ;

: SET-FRT-MODE-EXACT ( -- )  ['] FRT-DATUM= FRT-TEST=-XT ! ;
: SET-FRT-MODE-ABS   ( -- )  ['] FRT-ABS=   FRT-TEST=-XT ! ;
: SET-FRT-MODE-REL   ( -- )  ['] FRT-REL=   FRT-TEST=-XT ! ;
: SET-FRT-MODE-REL0  ( -- )  ['] FRT-REL0=  FRT-TEST=-XT ! ;

SET-FRT-MODE-EXACT


\ *** ARRAYS
(
The first array slot has index zero.
)

\ USER CONFIG DEFAULTS
[UNDEFINED] MAX-FRT-INITIALS [IF] 32 CONSTANT MAX-FRT-INITIALS [THEN]
[UNDEFINED] MAX-FRT-RESULTS  [IF] 32 CONSTANT MAX-FRT-RESULTS  [THEN]

\ Arrays begin with a cell-sized count.
CREATE frt-initials MAX-FRT-INITIALS 1+ CELLS ALLOT
CREATE frt-results  MAX-FRT-RESULTS  1+ CELLS ALLOT
CREATE frt-goals    MAX-FRT-RESULTS  1+ CELLS ALLOT

: 0FRT-A  ( +n 'array -- )  \ assume n<>0
(
Initialize array fr variables for garbage collection.
)
  swap 1+ cells over + swap DO 0 i ! 1 cells +LOOP ;

MAX-FRT-INITIALS frt-initials 0FRT-A
MAX-FRT-RESULTS  frt-results  0FRT-A
MAX-FRT-RESULTS  frt-goals    0FRT-A

\ Overflow checking to be done elsewhere.
: frt-a!   ( +n 'array -- ) (fr: f_[n-1] ... f_0 -- )
  2dup ! cell+ swap ( 'array n) dup
  IF   cells over + swap
       DO i fr! 1 cells +LOOP 
  ELSE 2drop THEN ;

: frt-a@   ( 'array -- +n ) (fr: -- f_[n-1] ... f_0 )
  dup @ dup >r ( 'array n r: n)
  IF   cell+  r@ 1- cells over + 
       DO i fr@ -1 cells +LOOP
  ELSE drop THEN r> ;

[UNDEFINED] NIP [IF] : NIP  ( a b -- b)  swap drop ; [THEN]

: frt-a<>  ( 'array1 'array2 -- [index+1]|-1|0 )
(
Nonzero return for errors:
  -1       unequal array size
  index+1  deepest unequal array elements
)
  dup @ swap cell+ >r >r  ( 'array1 r: 'f2 #left )
  dup @ r@ <> IF 2r> 2drop drop -1 EXIT THEN
  ( 'array1) dup cell+
  BEGIN  r@ ( 'array1 'f1 #left)
  WHILE  ( 'f1) dup fr@ cell+
         r> ( #left) 1- r>
         ( 'f2) dup fr@ cell+ >r ( #left) >r
         FRT-TEST=
  WHILE
  REPEAT ( 'array1 'f1) drop @ ( len) 2r> nip - ( index+1)
  ELSE   ( 'array1 'f1) 2drop 2r> 2drop 0
  THEN ;

\ *** TESTS

: FR{   ( <frinitials> -- )
(
Save the frstack <frinitials> for restoration by }FR. 
Clear <frresults> and <frgoals> storage to put it in a defined
state for }FR in case FR-> is missing.
)
  frdepth dup MAX-FRT-INITIALS > ABORT" TOO MANY FR INITIALS"
  frt-initials frt-a!   0 frt-results !   0 frt-goals ! ;

: FR->  ( <frresults> -- )
(
Record the depth and contents of the frstack.
)
  frdepth dup MAX-FRT-RESULTS > ABORT" TOO MANY FR RESULTS"
  frt-results frt-a! ;

: }FR   ( <frgoals> -- <frinitials> )
(
Save the frstack <frgoals>, compare them with the saved
<frresults>, and restore the <frinitials>.
)
  frdepth dup MAX-FRT-RESULTS > ABORT" TOO MANY FR GOALS"
  frt-goals frt-a!
  frt-results frt-goals frt-a<> dup frt-error-index !
  ?DUP IF -1 =
    IF   S" WRONG NUMBER OF FR RESULTS: "
    ELSE -1 frt-error-index +! S" INCORRECT FR RESULT: " THEN
    FRT-ERROR
  THEN
  frt-initials frt-a@ ( len) drop ;

BASE !

\ *** END of frtester.fs
