(    Title:  Tests for the GNU Multiplie Precision word set
      File:  gmpfr-test.fs
    Author:  David N. Williams
   Version:  0.8.0
   License:  Public Domain
  Log file:  gmp.log
   Started:  February 2, 2011
   Revised:  March 26, 2011
)
decimal

[UNDEFINED] \\ [IF]
  : \\  ( -- )  -1 parse 2drop BEGIN refill 0= UNTIL ;
[THEN]
: .#  ( n -- )  cr ." #" . cr ;

s" gmpfr.fs"	       INCLUDED
s" ttester-xf.fs"      INCLUDED
s" tester-display.fs"  INCLUDED
s" xtester-errors.fs"  INCLUDED
s" ftester-errors.fs"  INCLUDED
s" btester.fs"         INCLUDED
s" btester-errors.fs"  INCLUDED
true VERBOSE !

s" qtester.fs"         INCLUDED
s" qtester-errors.fs"  INCLUDED
s" mftester.fs"        INCLUDED
s" mftester-errors.fs" INCLUDED

[UNDEFINED] bits/cell [IF]
s" ADDRESS-UNIT-BITS" environment?
  0= [IF] cr .( ***Can't determine BITS/CELL) ABORT [THEN]
  1 cells * constant BITS/CELL [THEN]

BITS/CELL 32 = BITS/CELL 64 = or
0= [IF] cr .( *** BITS/CELL must be 32 or 64) ABORT [THEN]

s" MAX-N" environment?
0= [IF] cr .( ***Can't determine MAX-N) ABORT [THEN]
constant MAXN

s" MAX-U" environment?
0= [IF] cr .( ***Can't determine MAX-U) ABORT [THEN]
constant MAXU

-1 MAXU <>
[IF] cr .( ***System does not use two's complement) ABORT [THEN]

: ?.errors
  ?.xt-errors s"   " ?type
  ?.ft-errors s"   " ?type ?.cr
  ?.bt-errors s"   " ?type
  ?.qt-errors s"   " ?type ?.mft-errors ;

: ={  b{ q{ mf{ t{ ;    : =>  b-> q-> mf-> -> ;    : }=  }b }q }mf }t ;

\ before the first displayed line
PFE-HOST IFORTH-HOST or [IF] ?.cr [THEN]

?." GNU Multiple Precision word set tests
?." bits/limb: " mp_bits_per_limb ?. ?.cr


\ *** GENERAL USE

[UNDEFINED] S= [IF] : S=  ( s -- flag)  compare 0= ; [THEN] 

: MZSTR-s= ( s -- flag ) MZBUF mcount s= ;
: MZSTR.   ( -- )        MZBUF mcount type ;

\ *** INTEGERS

\ SETUP

bspace free throw
/MPZ 2 cells + 7 * CONSTANT /BTBUF
/BTBUF make-bspace

\ TROUBLE-SHOOTING WORDS

: .bstat  ( -- )
  VERBOSE @ IF
    ." bbuffer size: " bspace >/gbuf @ . ." bytes" cr
    ." bunused: " bspace gunused . ." bytes   " cr
    ." bdepth:  " bspace gdepth . cr
    base @ hex
    ." bbuffer: 0x" bspace >gbuf . cr
    ." bbreak:  0x" bspace >gbrkp @ . cr
    ." bsp:     0x" bspace >gsp @ . cr
    ." bsp0:    0x" bspace >gsp0 @ . cr
    bgarbage? IF ." There is some b garbage. "
    ELSE ." There is no b garbage. " THEN cr
    ." Garbage collection lock: "
    bgc-lock@ IF ." ON" ELSE ." OFF" THEN cr
    base ! THEN ;

: dump-bspace  ( -- )
  VERBOSE @ IF
    bspace bspace >/gbuf @ /gspace-header + cell+ ( #bytes) dump
    [ PFE-HOST ] [IF] cr [THEN]
    .bstat THEN ;

: dump-b  ( 'b -- )
  ." b: " bspace >/grec @ cell- bounds ?DO i @ . 1 cells +LOOP cr ;

: bsp@    ( --  bsp ) bspace >gsp @ ;
: 0bspace ( -- )      bspace 0gspace ;  \ leaves bvars in limbo

\ TESTS

TESTING  b" bs@ (bdrop) bdrop bdepth b. BCONSTANT BVARIABLE

: ?b.    (b: b -- ) VERBOSE @ IF b. ELSE bdrop THEN ;
: >mbstr ( 'b -- )  MZBUF cell+ base @ rot mpz_get_str zstrlen MZBUF ! ;

b"  0" bconstant  0.b
b"  1" bconstant  1.b   b"  -1" bconstant  -1.b
b"  2" bconstant  2.b   b"  -2" bconstant  -2.b
b"  3" bconstant  3.b   b"  -3" bconstant  -3.b
b"  4" bconstant  4.b   b"  -4" bconstant  -4.b 
b"  5" bconstant  5.b   b"  -5" bconstant  -5.b
b"  6" bconstant  6.b   b"  -6" bconstant  -6.b
b"  7" bconstant  7.b   b"  -7" bconstant  -7.b
b"  8" bconstant  8.b   b"  -8" bconstant  -8.b
b"  9" bconstant  9.b   b"  -9" bconstant  -9.b
b" 10" bconstant 10.b   b" -10" bconstant -10.b

0bspace
t{ b" 104" bdepth bs@ cell- @ ( blink) -> 1 bsp@ }t
t{ s" 104" MZSTR-s= -> true }t
t{ bs@ >mbstr s" 104" MZSTR-s= -> true }t
t{ bgarbage? bdepth BCONSTANT 104.b bgarbage? bdepth 104.b bdepth
   -> false 1 true 0 1 }t  \ copy grec to ext, garbage grec
?." You should see 104: " t{ ?b. bdepth -> 0 }t ?.cr

t{ : compiled-104.b  104.b ; -> }t
t{ compiled-104.b bs@ >mbstr s" 104" MZSTR-s= -> true }t

t{ 0.b bs@ >mbstr s" 0" MZSTR-s= -> true }t
t{ (bdrop) bdepth bdrop bdepth -> 1 0 }t

t{ BVARIABLE bvar bvar @ -> 0 }t

BITS/CELL 32 = hex
[IF]
  b"   7fffffff" bconstant MAXN.b
  b"   ffffffff" bconstant MAXU.b
  b" 1 fffffffe" bconstant 2*MAXU.b
[ELSE]
  b"   7fffffff ffffffff" bconstant MAXN.b
  b"   ffffffff ffffffff" bconstant MAXU.b
  b" 1 ffffffff fffffffe" bconstant 2*MAXU.b
[THEN]
decimal

BITS/CELL 32 = [IF]
t{   MAXN.b bs@ >mbstr s" 2147483647" MZSTR-s= bdrop -> true }t
t{ 2*MAXU.b bs@ >mbstr s" 8589934590" MZSTR-s= bdrop -> true }t
[ELSE]
t{   MAXN.b bs@ >mbstr  s" 9223372036854775807" MZSTR-s= bdrop -> true }t
t{ 2*MAXU.b bs@ >mbstr s" 36893488147419103230" MZSTR-s= bdrop -> true }t
[THEN]

TESTING  >bs >bs-copy b,b' bs2@ bs3@ b2drop b3drop

: bdynam? ( 'b -- flag ) bspace is-grec ; 

={ 1.b bs@ dup >bs bdynam?      => 1.b 1.b false }=
={ 1.b bs@ >bs-copy bs@ bdynam? => 1.b 1.b true }=

={ 2.b align here b,'b dup >bs = => true 2.b }=

={ 1.b 2.b     bs2@ >bs >bs     => 1.b 2.b 2.b 1.b }=
={ 1.b 2.b 3.b bs3@ >bs >bs >bs => 1.b 2.b 3.b 3.b 2.b 1.b }=

={ 1.b 2.b         b2drop => }=
={ 1.b 2.b 3.b     b2drop => 1.b }=
={ 1.b 2.b 3.b     b3drop =>  }=
={ 1.b 2.b 3.b 2.b b3drop => 1.b }=

TESTING  bsgn b> b= b< b<= b>= b0> b0= b0< b0>= b0<=

={  2.b bsgn =>  1 }=
={  0.b bsgn =>  0 }=
={ -1.b bsgn => -1 }=

t{ 0.b bdepth -> 1 }t      \ bstack convervation check...
={ 1.b 2.b b> => false }=
t{ bdrop bdepth -> 0 }t   \ ...for ={ => }=
={ 2.b 1.b b> => true }=
={ 2.b 2.b b> => false }=

={ 1.b 2.b b= => false }=
={ 2.b 1.b b= => false }=
={ 2.b 2.b b= => true }=

={ 1.b 2.b b< => true }=
={ 2.b 1.b b< => false }=
={ 2.b 2.b b< => false }=

={ 1.b 2.b b<= => true }=
={ 2.b 1.b b<= => false }=
={ 2.b 2.b b<= => true }=

={ 1.b 2.b b>= => false }=
={ 2.b 1.b b>= => true }=
={ 2.b 2.b b>= => true }=

={  1.b b0> => true }=
={  0.b b0> => false }=
={ -1.b b0> => false }=

={  1.b b0= => false }=
={  0.b b0= => true }=
={ -1.b b0= => false }=

={  1.b b0< => false }=
={  0.b b0< => false }=
={ -1.b b0< => true }=

={  1.b b0>= => true }=
={  0.b b0>= => true }=
={ -1.b b0>= => false }=

={  1.b b0<= => false }=
={  0.b b0<= => true }=
={ -1.b b0<= => true }=

TESTING  bn> bn= bn< bn>= bn<=

={  1.b 2 bn> => false }=
={  2.b 1 bn> => true }=
={  2.b 2 bn> => false }=

={  1.b 2 bn= => false }=
={  2.b 1 bn= => false }=
={  2.b 2 bn= => true }=

={  1.b 2 bn< => true }=
={  2.b 1 bn< => false }=
={  2.b 2 bn< => false }=

={  1.b 2 bn>= => false }=
={  2.b 1 bn>= => true }=
={  2.b 2 bn>= => true }=

={  1.b 2 bn<= => true }=
={  2.b 1 bn<= => false }=
={  2.b 2 bn<= => true }=

TESTING  bu> bu= bu< bu>= bu<=

={  2.b -1 bu> => false }=
={  2.b  1 bu> => true }=
={  2.b  2 bu> => false }=

={  1.b 2 bu= => false }=
={  2.b 1 bu= => false }=
={  2.b 2 bu= => true }=

={  1.b  2 bu< => true }=
={  2.b  1 bu< => false }=
={  2.b -1 bu< => true }=

={  1.b 2 bu>= => false }=
={  2.b 1 bu>= => true }=
={  2.b 2 bu>= => true }=

={  1.b -1 bu<= => true }=
={  2.b  1 bu<= => false }=
={  2.b  2 bu<= => true }=

TESTING  bdf> bdf= bdf< bdf>= bdf<=

={  1.b 2e bdf> => false }=
={  2.b 1e bdf> => true }=
={  2.b 2e bdf> => false }=

={  1.b 2e bdf= => false }=
={  2.b 1e bdf= => false }=
={  2.b 2e bdf= => true }=

={  1.b 2e bdf< => true }=
={  2.b 1e bdf< => false }=
={  2.b 2e bdf< => false }=

={  1.b 2e bdf>= => false }=
={  2.b 1e bdf>= => true }=
={  2.b 2e bdf>= => true }=

={  1.b 2e bdf<= => true }=
={  2.b 1e bdf<= => false }=
={  2.b 2e bdf<= => true }=

TESTING  b|>| b|=| b|<| b|<=| b|>=|

={  1.b -2.b b|>| => false }=
={  2.b -1.b b|>| => true }=
={ -2.b  2.b b|>| => false }=

={  1.b  2.b b|=| => false }=
={ -2.b  2.b b|=| => true }=
={  2.b -2.b b|=| => true }=

={  1.b -2.b b|<| => true }=
={  2.b  1.b b|<| => false }=
={  2.b -2.b b|<| => false }=

={  1.b  2.b b|<=| => true }=
={ -2.b  1.b b|<=| => false }=
={  2.b -2.b b|<=| => true }=

={  1.b  2.b b|>=| => false }=
={  2.b  1.b b|>=| => true }=
={ -2.b  2.b b|>=| => true }=

TESTING  bdf|>| bdf|=| bdf|<| bdf|<=| bdf|>=|

={  1.b -2e bdf|>| => false }=
={  2.b -1e bdf|>| => true }=
={ -2.b  2e bdf|>| => false }=

={  1.b  2e bdf|=| => false }=
={ -2.b  2e bdf|=| => true }=
={  2.b -2e bdf|=| => true }=

={  1.b -2e bdf|<| => true }=
={  2.b  1e bdf|<| => false }=
={  2.b -2e bdf|<| => false }=

={  1.b  2e bdf|<=| => true }=
={ -2.b  1e bdf|<=| => false }=
={  2.b -2e bdf|<=| => true }=

={  1.b  2e bdf|>=| => false }=
={  2.b  1e bdf|>=| => true }=
={ -2.b  2e bdf|>=| => true }=

TESTING  bu|>| bu|=| bu|<| bu|<=| bu|>=|

\ assume two's complement
={  2.b -1 bu|>| => false }=
={  2.b  1 bu|>| => true }=
={  2.b  2 bu|>| => false }=

={  1.b  2 bu|=| => false }=
={  2.b  1 bu|=| => false }=
={ -2.b  2 bu|=| => true }=

={  1.b  2 bu|<| => true }=
={ -2.b  1 bu|<| => false }=
={ -2.b -1 bu|<| => true }=

={  1.b 2 bu|>=| => false }=
={ -2.b 1 bu|>=| => true }=
={  2.b 2 bu|>=| => true }=

={  1.b -1 bu|<=| => true }=
={  2.b  1 bu|<=| => false }=
={  2.b  2 bu|<=| => true }=

TESTING  0ba 0by 0bz u>ba n>ba df>ba b>ba ba>b by>b bz>b ba.

={ 0ba ba>b => 0.b }=
={ 0by by>b => 0.b }=
={ 0bz bz>b => 0.b }=
hex
={ FFFFFFFF u>ba ba>b => b" ffffffff" }=
={ -1       n>ba ba>b => b" -1" }=
={ -2  n>ba ba>b => b" -2" }=
decimal
={ 13e    df>ba ba>b => b" 13" }=
={  2.b   b>ba ba>b =>  2.b }=

: ?ba. ( -- ) VERBOSE @ IF ba. THEN ;

?." You should see 105: " ={ b" 105" b>ba ?ba. => }= ?.cr

TESTING  ba>n  ba>u  ba>df  ba>df-exp

variable texp   0 texp !

={ MAXN.b b>ba ba>n  => MAXN }=
={ MAXU.b b>ba ba>u  => MAXU }=
={ 10.b   b>ba ba>df => 10e }=
\ 10.0 = 0.625 x 2^4
={ 10.b b>ba texp ba>df-exp texp @ => 0.625e 4 }=

TESTING  -ba |ba| ba+ by+ ba- by- bau+ bau-

={  3 n>ba -ba  ba>b => -3.b }=
={  3 n>ba |ba| ba>b =>  3.b }=
={ -2 n>ba |ba| ba>b =>  2.b }=

={ 0ba 1.b ba+ ba>b =>  1.b }=
={ -2.b    ba+ ba>b => -1.b }=
={ -3.b    ba- ba>b =>  2.b }=
hex
={ FFFFFFFF bau+ ba>b => b" 1 0000 0001" }=
={ FFFFFFFF bau- ba>b => b" 2" }=
decimal

TESTING  ba* ban* bau* ba*+ bau*+ ba*- bau*- ba2^u*

={  1 n>ba -2.b      ba*    ba>b => -2.b }=
={ -1                ban*   ba>b => 2.b }=
={ -1                bau*   ba>b => 2*MAXU.b }=
={ -1 n>ba  2.b  2.b ba*+   ba>b =>  3.b }=
={  1 n>ba -2.b  2   bau*+  ba>b => -3.b }=
={  1 n>ba  2.b  2.b ba*-   ba>b => -3.b }=
={ -1 n>ba -2.b  2   bau*-  ba>b =>  3.b }=
={  1 n>ba 2         ba2^u* ba>b =>  4.b }=

TESTING  bac/ bacrem bac/rem bacu/ bacurem bacu/rem bac/urem ba2^uc/ ba2^ucrem
---
The floored and symmetric division examples in the table below
are taken from DPANS94 3.2.2.1:

  Ceiled             Floored             Symmetric         
  num  den quo rem    num  den quo rem    num  den quo rem  
   10   7   2  -4      10   7   1   3      10   7   1   3   
  -10   7  -1  -3     -10   7  -2   4     -10   7  -1  -3   
   10  -7  -1   3      10  -7  -2  -4      10  -7  -1   3   
  -10  -7   2   4     -10  -7   1  -3     -10  -7   1  -3   

Note that we only need the first two rows to distinguish calls
among the c [ceiled], f [floored], and s [symmetric] functions.
---

={ 10.b b>ba 7.b bac/          ba>b =>  2.b }=
={ 10.b b>ba 7.b bacrem        ba>b => -4.b }=
={ 10.b b>ba 7.b bac/rem ba>b  by>b =>  2.b -4.b }=
={ 10.b b>ba 7   bauc/         ba>b =>  2.b  4 }=
={ 10.b b>ba 7   baucrem       ba>b => -4.b  4 }=
={ 10.b b>ba 7   bauc/rem ba>b by>b =>  2.b -4.b 4 }=
={ 10.b b>ba 7   bacurem       ba>b => 10.b  4 }=  \ no change in x
\ 2^3 = 8
={ 10.b b>ba 3 ba2^uc/	 ba>b =>  2.b }=
={ 10.b b>ba 3 ba2^ucrem ba>b => -6.b }=

={ -10.b b>ba 7.b bac/          ba>b => -1.b }=
={ -10.b b>ba 7.b bacrem        ba>b => -3.b }=
={ -10.b b>ba 7.b bac/rem ba>b  by>b => -1.b -3.b }=
={ -10.b b>ba 7   bauc/         ba>b => -1.b  3 }=
={ -10.b b>ba 7   baucrem       ba>b => -3.b  3 }=
={ -10.b b>ba 7   bauc/rem ba>b by>b => -1.b -3.b 3 }=
={ -10.b b>ba 7   bacurem       ba>b => -10.b 3 }=  \ no change in x
\ 2^3 = 8
={ -10.b b>ba 3 ba2^uc/	  ba>b => -1.b }=
={ -10.b b>ba 3 ba2^ucrem ba>b => -2.b }=

TESTING  baf/ bafrem baf/rem bafu/ bafurem bafu/rem baf/urem ba2^uf/ ba2^ufrem

={ 10.b b>ba 7.b baf/          ba>b => 1.b }=
={ 10.b b>ba 7.b bafrem        ba>b => 3.b }=
={ 10.b b>ba 7.b baf/rem ba>b  by>b => 1.b 3.b }=
={ 10.b b>ba 7   bauf/         ba>b => 1.b 3 }=
={ 10.b b>ba 7   baufrem       ba>b => 3.b 3 }=
={ 10.b b>ba 7   bauf/rem ba>b by>b => 1.b 3.b 3 }=
={ 10.b b>ba 7   bafurem       ba>b => 10.b 3 }=  \ no change in x
\ 2^3 = 8
={ 10.b b>ba 3 ba2^uf/	 ba>b => 1.b }=
={ 10.b b>ba 3 ba2^ufrem ba>b => 2.b }=

={ -10.b b>ba 7.b baf/          ba>b =>  -2.b }=
={ -10.b b>ba 7.b bafrem        ba>b =>   4.b }=
={ -10.b b>ba 7.b baf/rem ba>b  by>b =>  -2.b 4.b }=
={ -10.b b>ba 7   bauf/         ba>b =>  -2.b 4 }=
={ -10.b b>ba 7   baufrem       ba>b =>   4.b 4 }=
={ -10.b b>ba 7   bauf/rem ba>b by>b =>  -2.b 4.b 4 }=
={ -10.b b>ba 7   bafurem       ba>b => -10.b 4 }=  \ no change in x
\ 2^3 = 8
={ -10.b b>ba 3 ba2^uf/	  ba>b => -2.b }=
={ -10.b b>ba 3 ba2^ufrem ba>b =>  6.b }=

TESTING  bas/ basrem bas/rem basu/ basurem basu/rem bas/urem ba2^us/ ba2^usrem

={ 10.b b>ba 7.b bas/          ba>b => 1.b }=
={ 10.b b>ba 7.b basrem        ba>b => 3.b }=
={ 10.b b>ba 7.b bas/rem ba>b  by>b => 1.b 3.b }=
={ 10.b b>ba 7   baus/         ba>b => 1.b 3 }=
={ 10.b b>ba 7   bausrem       ba>b => 3.b 3 }=
={ 10.b b>ba 7   baus/rem ba>b by>b => 1.b 3.b 3 }=
={ 10.b b>ba 7   basurem       ba>b => 10.b 3 }=  \ no change in x
\ 2^3 = 8
={ 10.b b>ba 3 ba2^us/	 ba>b => 1.b }=
={ 10.b b>ba 3 ba2^usrem ba>b => 2.b }=

={ -10.b b>ba 7.b bas/          ba>b =>  -1.b }=
={ -10.b b>ba 7.b basrem        ba>b =>  -3.b }=
={ -10.b b>ba 7.b bas/rem ba>b by>b  =>  -1.b -3.b }=
={ -10.b b>ba 7   baus/         ba>b =>  -1.b  3 }=
={ -10.b b>ba 7   bausrem       ba>b =>  -3.b  3 }=
={ -10.b b>ba 7   baus/rem ba>b by>b =>  -1.b -3.b 3 }=
={ -10.b b>ba 7   basurem       ba>b => -10.b  3 }=  \ no change in x
\ 2^3 = 8
={ -10.b b>ba 3 ba2^us/	  ba>b => -1.b }=
={ -10.b b>ba 3 ba2^usrem ba>b => -2.b }=

TESTING  bamod baumod ba/exact bau/exact

={  10.b b>ba 7.b bamod ba>b  => 3.b }=
={ -10.b b>ba 7.b bamod ba>b  => 4.b }=
={  10.b b>ba 7   baumod ba>b => 3.b 3 }=
={ -10.b b>ba 7   baumod ba>b => 4.b 4 }=

={  10.b b>ba  2.b ba/exact ba>b  =>  5.b }=
={ -10.b b>ba  2.b ba/exact ba>b  => -5.b }=
={  10.b b>ba -2.b ba/exact ba>b  => -5.b }=
={  10.b b>ba  2   bau/exact ba>b =>  5.b }=
={ -10.b b>ba  2   bau/exact ba>b => -5.b }=

TESTING  ba^pow-mod ba^pow-mod-sec ba^u-mod ba^u u^p>ba

={  2.b b>ba 3.b 7.b ba^pow-mod ba>b     => 1.b }=
={  2.b b>ba 5.b 7.b ba^pow-mod ba>b     => 4.b }=
={  2.b b>ba 3.b 7.b ba^pow-mod-sec ba>b => 1.b }=
={  2.b b>ba 5.b 7.b ba^pow-mod-sec ba>b => 4.b }=
={  2.b b>ba 3   7.b ba^u-mod ba>b       => 1.b }=
={  2.b b>ba 5   7.b ba^u-mod ba>b       => 4.b }=

={ 2.b b>ba 3 ba^u   ba>b => 8.b }=
={ 2        3 u^p>ba ba>b => 8.b }=

TESTING  ba-nroot ba-nroot-rem ba-sqrt ba-sqrt-rem

={  8.b b>ba 3 ba-nroot     ba>b      => 2.b true }=
={ 10.b b>ba 3 ba-nroot     ba>b      => 2.b false }=
={  8.b b>ba 3 ba-nroot-rem ba>b by>b => 2.b 0.b }=
={ 10.b b>ba 3 ba-nroot-rem ba>b by>b => 2.b 2.b }=
={  9.b b>ba   ba-sqrt      ba>b      => 3.b }=
={ 10.b b>ba   ba-sqrt-rem  ba>b by>b => 3.b 1.b }=

TESTING  ba-next-prime ba-gcd bau-gcd ba-gcd-st ba-gcd-s ba-lcm  bau-lcm 

={ b" 258" b>ba        ba-next-prime ba>b           => b" 263" }=
={  b" 20" b>ba b" 12" ba-gcd        ba>b           => 4.b }=
={  b" 20" b>ba    12  bau-gcd       ba>b           => 4 4.b }=
={  b" 20" b>ba b" 12" ba-gcd-st     by>b bz>b ba>b => -1.b 2.b 4.b }=
={  b" 20" b>ba b" 12" ba-gcd-s      by>b ba>b      => -1.b 4.b }=
={  b" 20" b>ba b" 12" ba-lcm        ba>b           => b" 60" }=
={  b" 20" b>ba    12  bau-lcm       ba>b           => b" 60" }=

TESTING  ba-mod-invert ba-remove fac>ba ba-bin bin>ba

\ [(7 mod 4) x 3] mod 4 = 1
={     7.b b>ba 4.b ba-mod-invert ba>b => true 3.b }=
={ b" 20"  b>ba 2.b ba-remove ba>b => 2 5.b }=
={     4            fac>ba    ba>b => b" 24" }=
={   10.b  b>ba 3   ba-bin    ba>b => b" 120" }=
={   10         3   bin>ba    ba>b => b" 120" }=

TESTING  fib>ba fib2>ba luc>ba luc2>ba 

={ 6 fib>ba  ba>b      => 8.b }=
={ 6 fib2>ba by>b ba>b => 5.b 8.b }=
={ 6 luc>ba  ba>b      => b" 18" }=
={ 6 luc2>ba by>b ba>b => b" 11" b" 18" }=

TESTING  ba-and ba-or ba-xor ba-invert

hex
={ b" 0f0f" b>ba  b" f00" ba-and    ba>b => b" f00" }=
={ b" 0f0f" b>ba b" f000" ba-or     ba>b => b" ff0f" }=
={ b" 0f0f" b>ba  b" f00" ba-xor    ba>b => b" f" }=
={      1.b b>ba          ba-invert ba>b => -2.b }=
={   MAXU.b b>ba          ba-invert ba>b => MAXU.b 1 bu+ -b }=
decimal

TESTING  ba-bit-on ba-bit-off ba-bit-flip ba-bit-test

hex
={ b" 0f0f" b>ba 4 ba-bit-on   ba>b => b" f1f" }=
={ b" 0f0f" b>ba 3 ba-bit-off  ba>b => b" f07" }=
={ b" 0f0f" b>ba 3 ba-bit-flip ba>b => b" f07" }=
={ b" 0f07" b>ba 3 ba-bit-flip ba>b => b" f0f" }=
={ b" 0f0f" b>ba 3 ba-bit-test      => true }=
={ b" 0f0f" b>ba 4 ba-bit-test      => false }=
decimal

TESTING  n>b u>b df>b b>n b>u b>df b>df-exp

={ MAXN n>b => MAXN.b }=
={ MAXU u>b => MAXU.b }=
={ 10e df>b => 10.b }=

={ MAXN.b b>n => MAXN }=
={ MAXU.b b>u => MAXU }=
={ 10.b   b>df => 10e }=
\ 10.0 = 0.625 x 2^4
={ 10.b texp b>df-exp texp @ => 0.625e 4 }=

TESTING  -b |b| b+ b- bu+ bu- ub- b* bn* bu*

={  2.b     -b  => -2.b }=
={ -2.b     |b| =>  2.b }=
={ 1.b 2.b  b+  =>  3.b }=
={ 1.b 2.b  b-  => -1.b }=
={ 1.b 2    bu+ =>  3.b }=
={ 1.b 2    bu- => -1.b }=
={ 1   2.b  ub- => -1.b }=
={ 2.b 2.b  b*  =>  4.b }=
={ 2.b -2   bn* => -4.b }=
={ 2.b MAXU bu* => 2*MAXU.b }=

TESTING  b*+ b*- bu*+ bu*- b2^u*

={ -1.b  2.b  2.b b*+   =>  3.b }=
={  1.b -2.b  2   bu*+  => -3.b }=
={  1.b  2.b  2.b b*-   => -3.b }=
={ -1.b -2.b  2   bu*-  =>  3.b }=
={ -1.b  2        b2^u* => -4.b }=

TESTING  bc/ bcrem bc/rem bcu/ bcurem bcu/rem bc/urem b2^uc/ b2^ucrem

={ 10.b 7.b bc/     =>  2.b }=
={ 10.b 7.b bcrem   => -4.b }=
={ 10.b 7.b bc/rem  =>  2.b -4.b }=
={ 10.b 7   buc/    =>  2.b  4 }=
={ 10.b 7   bucrem  => -4.b  4 }=
={ 10.b 7   buc/rem =>  2.b -4.b 4 }=
={ 10.b 7   bcurem  =>  4 }=
\ 2^3 = 8
={ 10.b 3 b2^uc/   =>  2.b }=
={ 10.b 3 b2^ucrem => -6.b }=

={ -10.b 7.b bc/     => -1.b }=
={ -10.b 7.b bcrem   => -3.b }=
={ -10.b 7.b bc/rem  => -1.b -3.b }=
={ -10.b 7   buc/    => -1.b  3 }=
={ -10.b 7   bucrem  => -3.b  3 }=
={ -10.b 7   buc/rem => -1.b -3.b 3 }=
={ -10.b 7   bcurem  =>  3 }=
\ 2^3 = 8
={ -10.b 3 b2^uc/    => -1.b }=
={ -10.b 3 b2^ucrem  => -2.b }=

TESTING  bf/ bfrem bf/rem bfu/ bfurem bfu/rem bf/urem b2^uf/ b2^ufrem

={ 10.b 7.b bf/     => 1.b }=
={ 10.b 7.b bfrem   => 3.b }=
={ 10.b 7.b bf/rem  => 1.b 3.b }=
={ 10.b 7   buf/    => 1.b 3 }=
={ 10.b 7   bufrem  => 3.b 3 }=
={ 10.b 7   buf/rem => 1.b 3.b 3 }=
={ 10.b 7   bfurem  => 3 }=
\ 2^3 = 8
={ 10.b 3 b2^uf/    => 1.b }=
={ 10.b 3 b2^ufrem  => 2.b }=

={ -10.b 7.b bf/     => -2.b }=
={ -10.b 7.b bfrem   =>  4.b }=
={ -10.b 7.b bf/rem  => -2.b 4.b }=
={ -10.b 7   buf/    => -2.b 4 }=
={ -10.b 7   bufrem  =>  4.b 4 }=
={ -10.b 7   buf/rem => -2.b 4.b 4 }=
={ -10.b 7   bfurem  =>  4 }=
\ 2^3 = 8
={ -10.b 3 b2^uf/    => -2.b }=
={ -10.b 3 b2^ufrem  =>  6.b }=

TESTING  bs/ bsrem bs/rem bsu/ bsurem bsu/rem bs/urem b2^us/ b2^usrem

={ 10.b 7.b bs/     => 1.b }=
={ 10.b 7.b bsrem   => 3.b }=
={ 10.b 7.b bs/rem  => 1.b 3.b }=
={ 10.b 7   bus/    => 1.b 3 }=
={ 10.b 7   busrem  => 3.b 3 }=
={ 10.b 7   bus/rem => 1.b 3.b 3 }=
={ 10.b 7   bsurem  => 3 }=
\ 2^3 = 8
={ 10.b 3 b2^us/   => 1.b }=
={ 10.b 3 b2^usrem => 2.b }=

={ -10.b 7.b bs/     => -1.b }=
={ -10.b 7.b bsrem   => -3.b }=
={ -10.b 7.b bs/rem  => -1.b -3.b }=
={ -10.b 7   bus/    => -1.b  3 }=
={ -10.b 7   busrem  => -3.b  3 }=
={ -10.b 7   bus/rem => -1.b -3.b 3 }=
={ -10.b 7   bsurem  =>  3 }=
\ 2^3 = 8
={ -10.b 3   b2^us/   => -1.b }=
={ -10.b 3   b2^usrem => -2.b }=

TESTING  bmod bumod b/exact bu/exact

={  10.b 7.b bmod  => 3.b }=
={ -10.b 7.b bmod  => 4.b }=
={  10.b 7   bumod => 3.b 3 }=
={ -10.b 7   bumod => 4.b 4 }=

TESTING  bdivisible bu-divisible b2^u-divisible

={ 10.b 2.b bdivisible     => true }=
={  7.b 2.b bdivisible     => false }=
={ 10.b 2   bu-divisible   => true }=
={  7.b 2   bu-divisible   => false }=
={  4.b 2   b2^u-divisible => true }=
={  7.b 2   b2^u-divisible => false }=

TESTING  bcongruent bu-congruent b2^u-congruent

={  7.b 3.b 4.b bcongruent     => true }=
={ 10.b 3.b 4.b bcongruent     => false }=
={  7.b 3   4   bu-congruent   => true }=
={ 10.b 3   4   bu-congruent   => false }=
={  7.b 3.b 2   b2^u-congruent => true }=
={ 10.b 3.b 2   b2^u-congruent => false }=

TESTING  b^pow-mod b^pow-mod-sec b^u-mod b^u u^p>b

={  2.b 3.b 7.b b^pow-mod     => 1.b }=
={  2.b 5.b 7.b b^pow-mod     => 4.b }=
={  2.b 3.b 7.b b^pow-mod-sec => 1.b }=
={  2.b 5.b 7.b b^pow-mod-sec => 4.b }=
={  2.b 3   7.b b^u-mod       => 1.b }=
={  2.b 5   7.b b^u-mod       => 4.b }=

={ 2.b 3 b^u   => 8.b }=
={ 2   3 u^p>b => 8.b }=

TESTING  bnroot bnroot-rem bsqrt bsqrt-rem

={  8.b 3 bnroot     => 2.b true }=
={ 10.b 3 bnroot     => 2.b false }=
={  8.b 3 bnroot-rem => 2.b 0.b }=
={ 10.b 3 bnroot-rem => 2.b 2.b }=
={  9.b   bsqrt      => 3.b }=
={ 10.b   bsqrt-rem  => 3.b 1.b }=

TESTING  bperfect-pow bperfect-squ

={  8.b bperfect-pow => true }=
={ 10.b bperfect-pow => false }=
={  8.b bperfect-squ => false }=
={  9.b bperfect-squ => true }=

TESTING  bprob-prime bnext-prime bgcd bu-gcd>u bu-gcd bgcd-st bgcd-s

={ b" 258"     3  bprob-prime => 0 }=
={ b" 263"     3  bprob-prime => 2 }=
={ b" 258"        bnext-prime => b" 263" }=
={  b" 20" b" 12" bgcd        =>  4.b }=
={  b" 20"    12  bu-gcd>u    => 4  }=
={  b" 20"    12  bu-gcd      => 4 4.b }=
={  b" 20" b" 12" bgcd-st     => -1.b 2.b 4.b }=
={  b" 20" b" 12" bgcd-s      => -1.b 4.b }=

TESTING  blcm bu-lcm bmod-invert bjacobi blegendre bkronecker  
---
It's important to test returns of -1 for the jacobi, legendre,
and kronecker symbols, to be sure that's properly extended from
and int to a cell with 64-bit systems.  It's sufficient to do
that for one of bjacobi, blegendre, or bkronecker, because
mpz_jacobi, mpz_legendre, and mpz_kronecker are defined to be
identical in ligbmp 5.0.1, and for the n and u variants of
kronecker, because there no such variants for jacobi or
legendre.
---

={  b" 20" b" 12" blcm        => b" 60" }=
={  b" 20"    12  bu-lcm      => b" 60" }=
={     7.b    4.b bmod-invert => true 3.b }=

\ 20 mod 5 = 0
={ b" 20" b" 15" bjacobi => 0 }=

\ 100 = (2x5)^2 <> 0 mod 3
={ b" 100" 3.b blegendre => 1 }=

\ 101 <> x^2 <> 0 mod 3
={ b" 101" 3.b blegendre => -1 }=

\ 101 <> x^2 <> 0 mod 7
={ b" 101" 7.b blegendre => -1 }=

\ 21 = 3 x 7
={ b" 101" b" 21" bjacobi => -1 -1 * }=

\ gmp aliases kro(a,b) to jac(a,b), at least when b <> 2 
={  b" 20" b" 15" bkronecker => 0 }=
={ b" 101" b" 21" bkronecker => -1 -1 * }=

TESTING  bn-kronecker bu-kronecker nb-kronecker ub-kronecker

\ distinction between u and n not really tested

={ b"  -20" 15 bn-kronecker => 0 }=
={ b" -101" 21 bn-kronecker => -1 -1 * }=
={ b"  101"  3 bn-kronecker => -1 }=
={ b"   20" 15 bu-kronecker => 0 }=
={ b"  101" 21 bu-kronecker => -1 -1 * }=
={ b"  101"  3 bu-kronecker => -1 }=

={  -20 b" 15" nb-kronecker => 0 }=
={ -101 b" 21" nb-kronecker => -1 -1 * }=
={  101 b"  3" nb-kronecker => -1 }=
={   20 b" 15" ub-kronecker => 0 }=
={  101 b" 21" ub-kronecker => -1 -1 * }=
={  101 b"  3" ub-kronecker => -1 }=

TESTING  bremove fac>b bu-bin bin>b fib>b fib2>b luc>b luc2>b

\ [(7 mod 4) x 3] mod 4 = 1
={ b" 20"  2.b bremove => 2 5.b }=
={     4       fac>b   => b" 24" }=
={    10.b 3   bu-bin  => b" 120" }=
={    10   3   bin>b   => b" 120" }=

={ 6 fib>b  => 8.b }=
={ 6 fib2>b => 5.b 8.b }=
={ 6 luc>b  => b" 18" }=
={ 6 luc2>b => b" 11" b" 18" }=

TESTING  band bor bxor binvert

hex
={ b" 0f0f"  b" f00" band    =>  b" f00" }=
={ b" 0f0f" b" f000" bor     => b" ff0f" }=
={ b" 0f0f"  b" f00" bxor    =>    b" f" }=
={     1.b           binvert => -2.b }=
={  MAXU.b           binvert => MAXU.b 1 bu+ -b }=
decimal

TESTING  bpop-count bham-dist bscan0 bscan1

hex
={ b" 0f0f"          bpop-count =>  8 }=
={ b" 0f0f" b" 0e0e" bham-dist  =>  2 }=
={ b" 0f0f"       3  bscan0     =>  4 }=
={ b" 0f0f"       4  bscan1     =>  8 }=
decimal

TESTING  bbit-on bbit-off bbit-flip bbit-test

hex
={ b" 0f0f" 4 bbit-on   => b" f1f" }=
={ b" 0f0f" 3 bbit-off  => b" f07" }=
={ b" 0f0f" 3 bbit-flip => b" f07" }=
={ b" 0f07" 3 bbit-flip => b" f0f" }=
={ b" 0f0f" 3 bbit-test => true }=
={ b" 0f0f" 4 bbit-test => false }=
decimal

\ TESTING  bgarbage? collect-bgarbage

\ ={ bgarbage? collect-bgarbage bgarbage? => true true false }=


\ *** RATIONALS

\ SETUP

qspace free throw  \ release the default space
/MPQ 2 cells + 7 * CONSTANT /QTBUF
/QTBUF make-qspace

\ TROUBLE-SHOOTING WORDS

: .qstat  ( -- )
  VERBOSE @ IF
    ." qbuffer size: " qspace >/gbuf @ . ." bytes" cr
    ." qunused: " qspace gunused . ." bytes   " cr
    ." qdepth:  " qspace gdepth . cr
    base @ hex
    ." qbuffer: 0x" qspace >gbuf . cr
    ." qbreak:  0x" qspace >gbrkp @ . cr
    ." qsp:     0x" qspace >gsp @ . cr
    ." qsp0:    0x" qspace >gsp0 @ . cr
    decimal
    qgarbage? IF ." There is some q garbage. "
    ELSE ." There is no q garbage. " THEN cr
    ." Garbage collection lock: "
    qgc-lock@ IF ." ON" ELSE ." OFF" THEN cr
    base ! THEN ;

: dump-qspace  ( -- )
  VERBOSE @ IF
    qspace dup >/gbuf @ /gspace-header + cell+ ( #bytes) dump
    [ PFE-HOST ] [IF] cr [THEN]
    .qstat THEN ;

: dump-q  ( 'f -- )
  ." q: " qspace >/grec @ cell-  bounds ?DO i @ . 1 cells +LOOP cr ;

: qsp@    ( -- qsp ) qspace >gsp @ ;
: 0qspace ( -- )     qspace 0gspace ;

\ TESTS

TESTING  q" qs@ (qdrop) qdrop qdepth q. QCONSTANT QVARIABLE

: ?q.    (q: q -- ) VERBOSE @ IF q. ELSE qdrop THEN ;
: >mqstr ( 'b -- )  MZBUF cell+ base @ rot mpq_get_str zstrlen MZBUF ! ;

q" 0"   qconstant 0/1
q" 1/2" qconstant 1/2   q" -1/2" qconstant -1/2
q" 2/3" qconstant 2/3   q" -2/3" qconstant -2/3
q" 3/4" qconstant 3/4   q" -3/4" qconstant -3/4

0qspace
t{ q" 104/3" qdepth qs@ cell- @ ( blink) -> 1 qsp@ }t
t{ s" 104/3" MZSTR-s= -> true }t
t{ qs@ >mqstr s" 104/3" MZSTR-s= -> true }t
t{ qgarbage? qdepth QCONSTANT 104/3 qgarbage? qdepth 104/3 qdepth
   -> false 1 true 0 1 }t  \ copy grec to ext, garbage grec
?." You should see 104/3: " t{ ?q. qdepth -> 0 }t ?.cr

t{ : compiled-104/3  104/3 ; -> }t
t{ compiled-104/3 qs@ >mqstr s" 104/3" MZSTR-s= -> true }t

t{ 0/1 qs@ >mqstr s" 0" MZSTR-s= -> true }t
t{ (qdrop) qdepth qdrop qdepth -> 1 0 }t

t{ QVARIABLE qvar qvar @ -> 0 }t

t{ 1/2 qs@ >mqstr s" 1/2" MZSTR-s= qdrop -> true }t

TESTING  >qs >qs-copy q,q' qs2@ qs3@ q2drop q3drop

: qdynam? ( 'q -- flag ) qspace is-grec ; 

={ 1/2 qs@ dup >qs qdynam?      => 1/2 1/2 false }=
={ 1/2 qs@ >qs-copy qs@ qdynam? => 1/2 1/2 true }=

={ 2/3 align here q,'q dup >qs = => true 2/3 }=

={ 1/2 2/3     qs2@ >qs >qs     => 1/2 2/3 2/3 1/2 }=
={ 1/2 2/3 3/4 qs3@ >qs >qs >qs => 1/2 2/3 3/4 3/4 2/3 1/2 }=

={ 1/2 2/3         q2drop => }=
={ 1/2 2/3 3/4     q2drop => 1/2 }=
={ 1/2 2/3 3/4     q3drop =>  }=
={ 1/2 2/3 3/4 2/3 q3drop => 1/2 }=

TESTING  qsgn q> q= q< q<= q>= q0> q0= q0< q0>= q0<=

={  2/3 qsgn =>  1 }=
={  0/1 qsgn =>  0 }=
={ -1/2 qsgn => -1 }=

t{ 0/1 qdepth -> 1 }t     \ qstack convervation check...
={ 1/2 2/3 q> => false }=
t{ qdrop qdepth -> 0 }t   \ ...for ={ => }=
={ 2/3 1/2 q> => true }=
={ 2/3 2/3 q> => false }=

={ 1/2 2/3 q= => false }=
={ 2/3 1/2 q= => false }=
={ 2/3 2/3 q= => true }=

={ 1/2 2/3 q< => true }=
={ 2/3 1/2 q< => false }=
={ 2/3 2/3 q< => false }=

={ 1/2 2/3 q<= => true }=
={ 2/3 1/2 q<= => false }=
={ 2/3 2/3 q<= => true }=

={ 1/2 2/3 q>= => false }=
={ 2/3 1/2 q>= => true }=
={ 2/3 2/3 q>= => true }=

={  1/2 q0> => true }=
={  0/1 q0> => false }=
={ -1/2 q0> => false }=

={  1/2 q0= => false }=
={  0/1 q0= => true }=
={ -1/2 q0= => false }=

={  1/2 q0< => false }=
={  0/1 q0< => false }=
={ -1/2 q0< => true }=

={  1/2 q0>= => true }=
={  0/1 q0>= => true }=
={ -1/2 q0>= => false }=

={  1/2 q0<= => false }=
={  0/1 q0<= => true }=
={ -1/2 q0<= => true }=

TESTING  qnu> qnu= qnu< qnu>= qnu<=

={  1/2 2 3 qnu> => false }=
={  2/3 1 2 qnu> => true }=
={  2/3 2 3 qnu> => false }=

={  1/2 2 3 qnu= => false }=
={  2/3 1 3 qnu= => false }=
={  2/3 2 3 qnu= => true }=

={  1/2 2 3 qnu< => true }=
={  2/3 1 2 qnu< => false }=
={  2/3 2 3 qnu< => false }=

={  1/2 2 3 qnu>= => false }=
={  2/3 1 2 qnu>= => true }=
={  2/3 2 3 qnu>= => true }=

={  1/2 2 3 qnu<= => true }=
={  2/3 1 2 qnu<= => false }=
={  2/3 2 3 qnu<= => true }=

TESTING  quu> quu= quu< quu>= quu<=

={  2/3 MAXU 2 quu> => false }=
={  2/3    1 2 quu> => true }=
={  2/3    2 3 quu> => false }=

={  1/2 2 3 quu= => false }=
={  2/3 1 2 quu= => false }=
={  2/3 2 3 quu= => true }=

={  1/2    2 3 quu< => true }=
={  2/3    1 2 quu< => false }=
={  2/3 MAXU 2 quu< => true }=

={  1/2 2 3 quu>= => false }=
={  2/3 1 2 quu>= => true }=
={  2/3 2 3 quu>= => true }=

={  1/2 MAXU 2 quu<= => true }=
={  2/3    1 2 quu<= => false }=
={  2/3    2 3 quu<= => true }=

TESTING  0qa nu>qa uu>qa df>qa q>qa qa>q qa>df qacanon qa.

: ?qa. ( -- ) VERBOSE @ IF qa. THEN ; 

={ 0qa qa>q => 0/1 }=
hex
={ -1       2 nu>qa qa>q => q" -1/2" }=
={ FFFFFFFF 2 uu>qa qa>q => q" ffffffff/2" }=
decimal
={ 2 3 nu>qa qa>q =>     2/3 }=
={ 13e df>qa qa>q => q" 13/1" }=
={ 2/3 q>qa qa>q  =>     2/3 }=
={ 1/2 q>qa qa>df =>    0.5e }=

={  12 16 nu>qa qacanon qa>q => 3/4 }=

?." You should see 105/3: " ={ 105 3 uu>qa ?qa. => }= ?.cr

TESTING  -qa |qa| 1/qa qa+ qa- qa* qa/ qa2^u* qa2^u/

q" 4/3" qconstant 4/3

={  3 4 nu>qa -qa  qa>q => -3/4 }=
={  3 4 nu>qa |qa| qa>q =>  3/4 }=
={ -2 3 nu>qa |qa| qa>q =>  2/3 }=
={  3 4 nu>qa 1/qa qa>q =>  4/3 }=


={ 0qa 1/2 qa+  qa>q =>  1/2 }=
={ -2/3    qa+  qa>q => q" -1/6" }=
={ -3/4    qa-  qa>q => q"  7/12" }=

={  2 3 nu>qa 3/4 qa* qa>q => 1/2 }=
={  4 9 nu>qa 2/3 qa/ qa>q => 2/3 }=

={  2 3 nu>qa 3 qa2^u* qa>q => q" 16/3" }=
={  2 3 nu>qa 3 qa2^u/ qa>q => q" 1/12" }=

TESTING  nu>q uu>q df>q q>df qcanon

hex
={ -1       2 nu>q => q" -1/2" }=
={ FFFFFFFF 2 uu>q => q" ffffffff/2" }=
decimal
={ 2 3        nu>q =>     2/3 }=
={ 13e        df>q => q" 13/1" }=
={ q" 13/1"   q>df =>    13e }=

={  12 16 nu>q qcanon => 3/4 }=

TESTING  -q |q| 1/q q+ q- q* q/ q2^u* q2^u/

={  3/4 -q  => -3/4 }=
={  3/4 |q| =>  3/4 }=
={ -2/3 |q| =>  2/3 }=
={  3/4 1/q =>  4/3 }=

={    1/2    1/2  q+ =>   q" 1" }=
={   -3/4 q" 1/3" q- => q" -13/12" }=
={    2/3    3/4  q* =>      1/2 }=
={ q" 4/9"   2/3  q/ =>      2/3 }=

={  2/3 3 q2^u* => q" 16/3" }=
={  2/3 3 q2^u/ => q" 1/12" }=


\ *** FLOATING POINT

\ SETUP

mfspace free throw  \ release the default space
/MPF 2 cells + 7 * CONSTANT /MFTBUF
/MFTBUF make-mfspace

\ TROUBLE-SHOOTING WORDS

: .mfstat  ( -- )
  VERBOSE @ IF
    ." mfbuffer size: " mfspace >/gbuf @ . ." bytes" cr
    ." mfunused: " mfspace gunused . ." bytes   " cr
    ." mfdepth:  " mfspace gdepth . cr
    base @ hex
    ." mfbuffer: 0x" mfspace >gbuf . cr
    ." mfbreak:  0x" mfspace >gbrkp @ . cr
    ." mfsp:     0x" mfspace >gsp @ . cr
    ." mfsp0:    0x" mfspace >gsp0 @ . cr
    decimal
    ." faprec:   " fa-start-precision @ . cr
    mfgarbage? IF ." There is some mf garbage. "
    ELSE ." There is no mf garbage. " THEN cr
    ." Garbage collection lock: "
    mfgc-lock@ IF ." ON" ELSE ." OFF" THEN cr
    base ! THEN ;

: dump-mfspace  ( -- )
  VERBOSE @ IF
    mfspace dup >/gbuf @ /gspace-header + cell+ ( #bytes) dump
    [ PFE-HOST ] [IF] cr [THEN]
    mfspace .mfstat THEN ;

: dump-mf  ( 'f -- )
  ." mf: " mfspace >/grec @ cell-  bounds ?DO i @ . 1 cells +LOOP cr ;

: mfsp@    ( -- mfsp ) mfspace >gsp @ ;
: 0mfspace ( -- )      mfspace 0gspace ;

TESTING  set-default-mfprecision get-default-mfprecision

t{ get-default-mfprecision 256 set-default-mfprecision -> 128 }t
t{ get-default-mfprecision 128 set-default-mfprecision -> 256  }t
t{ get-default-mfprecision -> 128  }t

t{ fa-start-precision @ -> get-default-mfprecision }t

TESTING  mf" mfs@ (mfdrop) mfdrop mfdepth mf. MFCONSTANT MFVARIABLE

\ : ?mf.  (mf: f -- )  VERBOSE @ IF mf. ELSE mfdrop THEN ;
: ?mf.  (mf: f s: u -- )  VERBOSE @ IF mf. ELSE mfdrop drop THEN ;

mf"  0e0" mfconstant  0.f
mf"  1e0" mfconstant  1.f   mf"  -1e0" mfconstant  -1.f
mf"  2e0" mfconstant  2.f   mf"  -2e0" mfconstant  -2.f
mf"  3e0" mfconstant  3.f   mf"  -3e0" mfconstant  -3.f
mf"  4e0" mfconstant  4.f   mf"  -4e0" mfconstant  -4.f 
mf"  5e0" mfconstant  5.f   mf"  -5e0" mfconstant  -5.f
mf"  6e0" mfconstant  6.f   mf"  -6e0" mfconstant  -6.f
mf"  7e0" mfconstant  7.f   mf"  -7e0" mfconstant  -7.f
mf"  8e0" mfconstant  8.f   mf"  -8e0" mfconstant  -8.f
mf"  9e0" mfconstant  9.f   mf"  -9e0" mfconstant  -9.f
mf" 10e0" mfconstant 10.f   mf" -10e0" mfconstant -10.f

0mfspace
t{ mf" 104" mfdepth mfs@ cell- @ ( blink) -> 1 mfsp@ }t
t{ s" 104" MZSTR-s= -> true }t
\ t{ mfs@ >MFSTR s" 104" MZSTR-s= MFEXP @ -> true 3 }t
t{ 3 mfs@ >MFSTR s" 104" MZSTR-s= MFEXP @ -> true 3 }t
t{ mfgarbage? mfdepth MFCONSTANT 104.f mfgarbage? mfdepth 104.f mfdepth
   -> false 1 true 0 1 }t  \ copy grec to ext, garbage grec
?." You should see 0.104E3: " t{ 3 ?mf. mfdepth -> 0 }t ?.cr

t{ : compiled-104.f  104.f ; -> }t
\ t{ MFEXP off compiled-104.f mfs@ >MFSTR s" 104" MZSTR-s= MFEXP @ -> true 3 }t
t{ MFEXP off compiled-104.f 3 mfs@ >MFSTR s" 104" MZSTR-s= MFEXP @ -> true 3 }t

\ t{ 0.f mfs@ >MFSTR s" " MZSTR-s= -> true }t  \ no trailing zero digits
t{ 1 0.f mfs@ >MFSTR s" " MZSTR-s= -> true }t  \ no trailing zero digits
t{ (mfdrop) mfdepth mfdrop mfdepth -> 1 0 }t

t{ FVARIABLE fvar fvar @ -> 0 }t

TESTING  >mfs >mfs-copy mf,f' mfs2@ mfs3@ mf2drop mf3drop

: mfdynam? ( 'f -- flag ) mfspace is-grec ; 

={ 1.f mfs@ dup >mfs mfdynam?       => 1.f 1.f false }=
={ 1.f mfs@ >mfs-copy mfs@ mfdynam? => 1.f 1.f true }=

={ 2.f align here mf,'f dup >mfs = => true 2.f }=

={ 1.f 2.f     mfs2@ >mfs >mfs      => 1.f 2.f 2.f 1.f }=
={ 1.f 2.f 3.f mfs3@ >mfs >mfs >mfs => 1.f 2.f 3.f 3.f 2.f 1.f }=

={ 1.f 2.f         mf2drop => }=
={ 1.f 2.f 3.f     mf2drop => 1.f }=
={ 1.f 2.f 3.f     mf3drop =>  }=
={ 1.f 2.f 3.f 2.f mf3drop => 1.f }=

TESTING  mfsgn mf> mf= mf< mf<= mf>= mf0> mf0= mf0< mf0>= mf0<=

={  2.f mfsgn =>  1 }=
={  0.f mfsgn =>  0 }=
={ -1.f mfsgn => -1 }=

t{ 0.f mfdepth -> 1 }t      \ mfstack convervation check...
={ 1.f 2.f mf> => false }=
t{ mfdrop mfdepth -> 0 }t   \ ...for ={ => }=
={ 2.f 1.f mf> => true }=
={ 2.f 2.f mf> => false }=

={ 1.f 2.f mf= => false }=
={ 2.f 1.f mf= => false }=
={ 2.f 2.f mf= => true }=

={ 1.f 2.f mf< => true }=
={ 2.f 1.f mf< => false }=
={ 2.f 2.f mf< => false }=

={ 1.f 2.f mf<= => true }=
={ 2.f 1.f mf<= => false }=
={ 2.f 2.f mf<= => true }=

={ 1.f 2.f mf>= => false }=
={ 2.f 1.f mf>= => true }=
={ 2.f 2.f mf>= => true }=

={  1.f mf0> => true }=
={  0.f mf0> => false }=
={ -1.f mf0> => false }=

={  1.f mf0= => false }=
={  0.f mf0= => true }=
={ -1.f mf0= => false }=

={  1.f mf0< => false }=
={  0.f mf0< => false }=
={ -1.f mf0< => true }=

={  1.f mf0>= => true }=
={  0.f mf0>= => true }=
={ -1.f mf0>= => false }=

={  1.f mf0<= => false }=
={  0.f mf0<= => true }=
={ -1.f mf0<= => true }=

TESTING  mfn> mfn= mfn< mfn>= mfn<=

={  1.f 2 mfn> => false }=
={  2.f 1 mfn> => true }=
={  2.f 2 mfn> => false }=

={  1.f 2 mfn= => false }=
={  2.f 1 mfn= => false }=
={  2.f 2 mfn= => true }=

={  1.f 2 mfn< => true }=
={  2.f 1 mfn< => false }=
={  2.f 2 mfn< => false }=

={  1.f 2 mfn>= => false }=
={  2.f 1 mfn>= => true }=
={  2.f 2 mfn>= => true }=

={  1.f 2 mfn<= => true }=
={  2.f 1 mfn<= => false }=
={  2.f 2 mfn<= => true }=

TESTING  mfu> mfu= mfu< mfu>= mfu<=

={  2.f -1 mfu> => false }=
={  2.f  1 mfu> => true }=
={  2.f  2 mfu> => false }=

={  1.f 2 mfu= => false }=
={  2.f 1 mfu= => false }=
={  2.f 2 mfu= => true }=

={  1.f  2 mfu< => true }=
={  2.f  1 mfu< => false }=
={  2.f -1 mfu< => true }=

={  1.f 2 mfu>= => false }=
={  2.f 1 mfu>= => true }=
={  2.f 2 mfu>= => true }=

={  1.f -1 mfu<= => true }=
={  2.f  1 mfu<= => false }=
={  2.f  2 mfu<= => true }=

TESTING  mfdf> mfdf= mfdf< mfdf>= mfdf<=

={  1.f 2e mfdf> => false }=
={  2.f 1e mfdf> => true }=
={  2.f 2e mfdf> => false }=

={  1.f 2e mfdf= => false }=
={  2.f 1e mfdf= => false }=
={  2.f 2e mfdf= => true }=

={  1.f 2e mfdf< => true }=
={  2.f 1e mfdf< => false }=
={  2.f 2e mfdf< => false }=

={  1.f 2e mfdf>= => false }=
={  2.f 1e mfdf>= => true }=
={  2.f 2e mfdf>= => true }=

={  1.f 2e mfdf<= => true }=
={  2.f 1e mfdf<= => false }=
={  2.f 2e mfdf<= => true }=

TESTING  0fa n>fa u>fa df>fa mf>fa fa>mf fa>n fa>u fa>df fa>df-exp fa.

: ?fa. ( u -- ) VERBOSE @ IF fa. ELSE drop THEN ;

BITS/CELL 32 = hex
[IF]
  mf" 7fff ffff"   mf" ffff ffff"
[ELSE]
  mf" 7fff ffff ffff ffff"   mf" ffff ffff ffff ffff"
[THEN] decimal
mfconstant MAXU.f   mfconstant MAXN.f

={ 0fa         fa>mf => 0.f }=
={ MAXN n>fa   fa>mf => MAXN.f }=
={ MAXU u>fa   fa>mf => MAXU.f }=
={  2e  df>fa  fa>mf => 2.f }=
={  2.f mf>fa fa>mf => 2.f }=

={ -10.f mf>fa fa>n  => -10 }=
={ -10.f mf>fa fa>u  =>  10 }=
={  10.f mf>fa fa>df =>  10e }=
\ 10.0 = 0.625 x 2^4
={ 10.f mf>fa texp fa>df-exp texp @ => 0.625e 4 }=

?." You should see 0.151@3: " 105 u>fa 8 base ! 3 ?fa. decimal ?.cr

TESTING  -fa |fa| fa+ fa- fau+ fau- ufa- fa* fau* fa/ fau/ ufa/

={  3 n>fa -fa  fa>mf => -3.f }=
={  3 n>fa |fa| fa>mf =>  3.f }=
={ -2 n>fa |fa| fa>mf =>  2.f }=
={ 0fa 1.f fa+  fa>mf =>  1.f }=
={ -2.f    fa+  fa>mf => -1.f }=
={ -3.f    fa-  fa>mf =>  2.f }=
hex
={ FFFFFFFF fau+ fa>mf => mf" 1 0000 0001" }=
={ FFFFFFFF fau- fa>mf => mf" 2" }=
decimal
={  2 3 n>fa   ufa-  fa>mf =>  -1.f }=

={  2 n>fa 4.f fa*  fa>mf => 8.f }=
={  3 n>fa 2   fau* fa>mf => 6.f }=
={  8 n>fa 2.f fa/  fa>mf => 4.f }=
={ 10 n>fa 2   fau/ fa>mf => 5.f }=
={ 12 2 n>fa   ufa/ fa>mf => 6.f }=

TESTING  fasqrt usqrt>fa fa^u fa2^u* fa/2^u farel-

={ 16 n>fa     fasqrt fa>mf => 4.f }=
={ 25        usqrt>fa fa>mf => 5.f }=
={  4 n>fa 2   fa^u   fa>mf => mf" 16" }=
={  4 n>fa 3   fa2^u* fa>mf => mf" 32" }=
={ 16 n>fa 2   fa/2^u fa>mf => 4.f }=
={ 16 n>fa 2.f farel- fa>mf => 7.f }=

TESTING  fa-ceil fa-floor fa-trunc fa-integer

={  1.5e df>fa fa-ceil  fa>mf =>  2.f }=
={ -1.5e df>fa fa-floor fa>mf => -2.f }=
={ -1.5e df>fa fa-trunc fa>mf => -1.f }=

={  1.5e df>fa fa-integer => false }=
={  2.0e df>fa fa-integer => true }=


TESTING  n>mf u>mf df>mf mf>n mf>u mf>df mf>df-exp

={ MAXN n>mf => MAXN.f }=
={ MAXU u>mf => MAXU.f }=
={ 10e df>mf => 10.f }=

={ MAXN.f mf>n  => MAXN }=
={ MAXU.f mf>u  => MAXU }=
={ 10.f   mf>df => 10e }=
\ 10.0 = 0.625 x 2^4
={ 10.f texp mf>df-exp texp @ => 0.625e  4 }=

TESTING  -mf |mf| mf+ mf- mfu+ mfu- umf- mf* mfu* mf/ mfu/ umf/

={  3 n>mf -mf    => -3.f }=
={  3 n>mf |mf|   =>  3.f }=
={ -2 n>mf |mf|   =>  2.f }=
={ -2.f  1.f mf+  => -1.f }=
={  5.f -3.f mf-  =>  8.f }=
={ 1.f MAXU  mfu+ => MAXU.f 1.f mf+ }=
={ MAXU.f 1  mfu- => MAXU.f 1.f mf- }=
={  2    3.f umf- =>  -1.f }=

={  2 n>mf 4.f mf*  => 8.f }=
={  3 n>mf 2   mfu* => 6.f }=
={  8 n>mf 2.f mf/  => 4.f }=
={ 10 n>mf 2   mfu/ => 5.f }=
={ 12 2 n>mf   umf/ => 6.f }=

TESTING  mfsqrt usqrt>mf mf^u mf*2^u mf/2^u mfrel-

={ 16 n>mf    mfsqrt  => 4.f }=
={ 25       usqrt>mf  => 5.f }=
={  4 n>mf 2   mf^u   => mf" 16" }=
={  4 n>mf 3   mf*2^u => mf" 32" }=
={ 16 n>mf 2   mf/2^u => 4.f }=
={ 16 n>mf 2.f mfrel- => 7.f }=

TESTING  mf-ceil mf-floor mf-trunc mf-integer

={ mf"  1.5" mf-ceil  =>  2.f }=
={ mf" -1.5" mf-floor => -2.f }=
={ mf" -1.5" mf-trunc => -1.f }=

={ mf" 1.5" mf-integer => false }=
={     2.f  mf-integer => true }=


\ *** MIXED STACK

TESTING  q>b mf>b b>q mf>q b>mf q>mf

 q" 3/1" qconstant 3/1   q" 4/1" qconstant 4/1   q" 5/1" qconstant 5/1
 q" 3/5" qconstant 3/5   q" 4/5" qconstant 4/5
 q" 9/4" qconstant 9/4
mf" 0.5" mfconstant 0.5.f

={ 9/4 q>b  => 2.b }=
={ 3.f mf>b => 3.b }=
={ 4.b b>q  => 4/1 }=
={ 5.f mf>q => 5/1 }=
={ 2.b b>mf => 2.f }=
={ 1/2 q>mf => 0.5.f }=

TESTING  qa>b fa>b ba>q fa>q ba>mf qa>mf

={ 9 4 uu>qa qa>b  => 2.b }=
={ 3   u>fa  fa>b  => 3.b }=
={ 3   u>ba  ba>q  => 3/1 }=
={ 5   u>fa  fa>q  => 5/1 }=
={ 2   u>ba  ba>mf => 2.f }=
={ 1 2 uu>qa qa>mf => 0.5.f }=

TESTING  q>ba mf>ba b>qa mf>qa b>fa q>fa

={ 9/4 q>ba  ba>b  => 2.b }=
={ 3.f mf>ba ba>b  => 3.b }=
={ 4.b b>qa  qa>q  => 4/1 }=
={ 5.f mf>qa qa>q  => 5/1 }=
={ 2.b b>fa  fa>mf => 2.f }=
={ 1/2 q>fa  fa>mf => 0.5.f }=

TESTING  qa>ba fa>ba ba>qa fa>qa ba>fa qa>fa

={ 9 4 uu>qa qa>ba ba>b  => 2.b }=
={ 3e  df>fa fa>ba ba>b  => 3.b }=
={ 4   u>ba  ba>qa qa>q  => 4/1 }=
={ 5e  df>fa fa>qa qa>q  => 5/1 }=
={ 2   n>ba  ba>fa fa>mf => 2.f }=
={ 1 2 nu>qa qa>fa fa>mf => 0.5.f }=

TESTING  qnum>b qden>b qanum>b qaden>b b>qnum b>qden b>qanum b>qaden
---
These tests also confirm the expected behavior for noncanonical
q's, which proves that the tests of QCANON and QACANON above are
not empty.
---
={ 12 16 nu>q  qnum>b  => b" 12" }=
={ 12 16 uu>q  qden>b  => b" 16" }=
={ 12 16 nu>qa qanum>b => b" 12" }=
={ 12 16 uu>qa qaden>b => b" 16" }=

={ 3.b 4/5       b>qnum       => 3/5 }=
={ 3.b 4/5       b>qden       => 4/3 }=
={ 3.b 4 5 nu>qa b>qanum qa>q => 3/5 }=
={ 3.b 4 5 nu>qa b>qaden qa>q => 4/3 }=

TESTING  qnum>ba qden>ba qanum>ba qaden>ba ba>qnum ba>qden ba>qanum ba>qaden

={  3  4 nu>q  qnum>ba  ba>b => 3.b }=
={  3  4 nu>q  qden>ba  ba>b => 4.b }=
={  3  4 nu>qa qanum>ba ba>b => 3.b }=
={  3  4 nu>qa qaden>ba ba>b => 4.b }=

={ 3 n>ba 4/5       ba>qnum        => 3/5 }=
={ 3 n>ba 4/5       ba>qden        => 4/3 }=
={ 3 u>ba 4 5 uu>qa ba>qanum qa>q => 3/5 }=
={ 3 u>ba 4 5 uu>qa ba>qaden qa>q => 4/3 }=

TESTING  RANDSTATE: free-randstate free-randstates
---
Stack effect only.
---

\ Should do nothing.
={ free-randstates => }=

={ RANDSTATE: 'trand align 'trand /GMP-RANDSTATE + cell+ aligned => here }=
={ 'trand /GMP-RANDSTATE aligned + @ => 0 }=  \ first in list

TESTING  randinit-default randinit-mt randinit-lc-2^u randinit-lc-bits
---
Stack effect only.  Although the fact that CLEAR-RANDSTATE in
the following doesn't give a malloc double-free error indicates
that initialization happend.
---

={          'trand randinit-default => 'trand free-randstate }=
={ 3.b 5 64 'trand randinit-lc-2^u  => 'trand free-randstate }=
={ 128      'trand randinit-lc-bits => true 'trand free-randstate }=
={ 129      'trand randinit-lc-bits => false }= \ no clear, wasn't initialized

\ Mersenne twister for remaining tests.
={          'trand randinit-mt      => 'trand free-randstate }=

TESTING  b>seed u>seed rand-#bits>u rand[0,n-1]>u

={ 'trand randinit-mt b" 93" 'trand b>seed => 'trand free-randstate }=
={ 'trand randinit-mt    93  'trand u>seed => }=

: reseed  ( -- )  93 'trand u>seed ;

: ?.rand-32bits ( -- )
  VERBOSE @ IF 4 0 DO 32 'trand rand-#bits>u u. LOOP THEN ;

?." 32-bit randoms: " ={ ?.rand-32bits => }= ?.cr

: ?.rand<100 ( -- )
  VERBOSE @ IF 4 0 DO 100 'trand rand[0,n-1]>u u. LOOP THEN ;

={ b" 93" 'trand b>seed => }=
?." randoms < 100 reseeded with b = 93: " ={ ?.rand<100 => }= ?.cr

={ reseed => }=
?." randoms < 100 reseeded with u = 93: " ={ ?.rand<100 => }= ?.cr

TESTING  rand-#bits>ba rand2-#bits>ba rand-#bits>b rand2-#bits>b

32 constant #tbits

: "rand-32bits>ba."   s" #tbits 'trand rand-#bits>ba ?ba. space" ;
: "rand-32bits>b."    s" #tbits 'trand rand-#bits>b ?b. space" ;
: "rand2-32bits>ba."  s" #tbits 'trand rand2-#bits>ba ?ba. space" ;
: "rand2-32bits>b."   s" #tbits 'trand rand2-#bits>b ?b. space" ;

\ RESEED has no effect on the limbs versions, which do not use
\ 'trand
: ?.rands  ( s -- )
  VERBOSE @ IF reseed 4 0 DO 2dup evaluate LOOP THEN 2drop ;

?." You should see 2 identical + 2 identical lines of random integers:

hex
={ "rand-32bits>ba."  ?.rands => }= ?.cr
={ "rand-32bits>b."   ?.rands => }= ?.cr
={ "rand2-32bits>ba." ?.rands => }= ?.cr
={ "rand2-32bits>b."  ?.rands => }= ?.cr
decimal

TESTING  rand-sbits>fa rand2-limbs>fa rand-sbits>mf rand2-limbs>mf

: "rand-32bits>fa."  s" #tbits 'trand rand-sbits>fa 8 ?fa." ;
: "rand-32bits>mf."  s" #tbits 'trand rand-sbits>mf 8 ?mf." ;
: "rand2-limbs>fa."  s" 1 1 rand2-limbs>fa 8 ?fa." ;
: "rand2-limbs>mf."  s" 1 1 rand2-limbs>mf 8 ?mf." ;

?." You should see 2 identical and 2 different lines of random hex floats:
hex
={ "rand-32bits>fa." ?.rands => }= ?.cr
={ "rand-32bits>mf." ?.rands => }= ?.cr
={ "rand2-limbs>fa." ?.rands => }= ?.cr
={ "rand2-limbs>mf." ?.rands => }= ?.cr
decimal

TESTING  copy-rand free-randstates

t{ create 'trand2 /GMP-RANDSTATE allot -> }t
={ 'trand 'trand2 copy-rand => }=

: ?.rand2-32bits ( -- )
  VERBOSE @ IF 4 0 DO 32 'trand2 rand-#bits>u u. LOOP THEN ;

: "rand2-32bits>b."    s" #tbits 'trand2 rand-#bits>b ?b. space" ;

?." You should see 2 identical lines of random integers:

={ "rand-32bits>b."  ?.rands => }= ?.cr
={ "rand2-32bits>b." ?.rands => }= ?.cr

\ stack effect only
={ free-randstates 'trand2 free-randstate => }=

IFORTH-HOST 0= [IF]
TESTING  exceptions

\ enough for 4 input characters or 3 output characters with
\ possible sign (b, q, or \ mf signicand)
1 cells 5 + to /MZBUF

={ /MZBUF 1+ ' ?MZBUF CATCH nip => -2 }=
={ b" 1234" => b" 1234" }=

?." You should see an integer and a rational:

={  123 n>b ' ?b. CATCH => 0 }=
t{ 1234 n>b ' b.  CATCH -> -2 }t
\ Uncomment to see exception.
\ b.
bdrop

={  1 5 nu>q ' ?q. CATCH => 0 ?.cr }=
t{ 12 5 nu>q '  q. CATCH -> -2 }t
\ Uncomment to see exception.
\ q.
qdrop

={ 123 u>fa 3 fa ' >MFSTR CATCH         => 0 }=
={ 123 u>fa 4 fa ' >MFSTR CATCH nip nip => -2 }=

\ Uncomment a line to see an exception.
\ b" 1q"
\ q" 1:2"
\ mf" 38p3"
[THEN]

\ cleanup
free-brecs   BSPACE free throw
free-qrecs   QSPACE free throw
free-mfrecs MFSPACE free throw
free-bconstants free-qconstants free-mfconstants

?.errors GFORTH-HOST [IF] ?.cr [THEN]
