(     Title:  GNU Multiple Precision word list
       File:  gmpfr.fs
     Author:  David N. Williams
    Version:  0.8.0
    License:  LGPL
  Test file:  gmpfr-test.fs
   Log file:  gmpfr.log
    Started:  February 2, 2011
    Revised:  March 26, 2011
)
\ Copyright (C) 2011 David N. Williams
(
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or at your option any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
MA 02111-1307 USA.

If you take advantage of the option in the LGPL to put a
particular version of this library under the GPL, the author[s]
would regard it as polite if you would put any direct
modifications under the LGPL as well, and include a copy of this
request near the beginning of the modified library source.  A
"direct modification" is one that enhances or extends the
library in line with its original concept, as opposed to
developing a distinct application or library which might use it.

This file defines Forth words for gmp big integers, rationals,
and floats, designated b, q, and mf for short, in terms of
bindings for the gmp library, libgmp, and the garbage-collected
record stack library, gcrstack.fs.

Currently the documentation of word actions is skimpy or
nonexistent.  It had been hoped that the rather well written GMP
library documentation could serve that purpose, but now we know
better.  It is nevertheless indispensable:

  http://gmplib.org/manual/
  http://gmplib.org/gmp-man-5.0.1.pdf

The test file gmpfr-test.fs contains many syntax examples.

Garbage collection is supposed to be transparent to the user.
See, however, the warning in gcrstack.fs about words that fetch
references from the record stack to the data stack.  It applies
to the stack fetch words in this library.  Generally speaking,
such references should appear on the data stack only under the
hood.

The current design puts a further restriction on the use of such
words, due to the immediate freeing of the memory managed by
libgmp records when they are garbaged.  Operations that could
result in garbaging must be postponed until after libgmp
functions that use the record data are finished.  Again, such
operations are expected to occur only under the hood.
)

s" FORTH-NAME" ENVIRONMENT? dup [IF]
  drop s" pfe" compare 0=
  0 REDEFINED-MSG !
[THEN] CONSTANT PFE-HOST

s" gforth" environment? dup [IF]
  ( s flag) nip nip
  0 WARNINGS !
[THEN] CONSTANT GFORTH-HOST

s" IFORTH" environment? dup [IF]
  ( flag) drop
  0 WARNING !
[THEN] CONSTANT IFORTH-HOST

PFE-HOST    [IF] LOADM libgmp-pfe  LOADM zchar [THEN]
GFORTH-HOST [IF] s" libgmp-gforth.fs"  REQUIRED [THEN]
IFORTH-HOST [IF] s" libgmp-iforth.frt" INCLUDED [THEN]

[UNDEFINED] /MPZ [IF] 12 constant /MPZ [THEN]

IFORTH-HOST [IF]
s" gcrstack.fs" INCLUDED
s" mstrings.fs" INCLUDED  \ loads parsing.fs
[ELSE]
s" gcrstack.fs" REQUIRED
s" mstrings.fs" REQUIRED  \ loads parsing.fs
[THEN]

: GMP-STACK-LIBRARY  s" GMP stack library version 0.7.1" ;


\ *** TABLE OF CONTENTS
(
  0 NOTATION AND TERMINOLOGY

  1 GENERAL SUPPORT
      MZBUF  /MZBUF  MZ"
      DEFAULT-MZBUF  DEFAULT-/MZBUF

  2 INTEGERS

  2.1 RECORD SPACE
      DEFAULT-/BBUF  BSPACE  make-bspace  free-brecs

  2.2 GARBAGE COLLECTION
      collect-bgarbage  bgarbage?
      bgc-off  bgc-on  bgc-lock@  bgc-lock!

  2.3 STACK
      (b:  bunused  bdepth  bs@  bs2@  bs3@  >bs  >bs-copy
      BVARIABLE  BCONSTANT  free-bconstants
      b"  b@  b!  bdrop  b2drop  b3drop  bdup  b2dup  bswap
      bnip  bover  btuck  bpick  b.

  2.4 ACCUMULATOR ARITHMETIC
      b>ba   ba>b   by>b   bz>b  0ba  0by  0bz
      n>ba   u>ba   df>ba
      ba>n   ba>u   ba>df  ba>df-exp  ba.

      \ elementary
      -ba  |ba|
      ba+  ba-   bau+  bau-  uba-
      ba*  ban*  bau*  ba*+  bau*+  ba*-  bau*-
      ba2^u*

      \ ceiled division
      bac/   bacrem   bac/rem
      bacu/  bacurem  bacu/rem  bac/urem  ba2^uc/  ba2^ucrem

      \ floored division
      baf/   bafrem   baf/rem
      bafu/  bafurem  bafu/rem  baf/urem  ba2^uf/  ba2^ufrem

      \ symmetric division
      bas/   basrem   bas/rem
      basu/  basurem  basu/rem  bas/urem  ba2^us/  ba2^usrem

      \ exact division
      ba/exact  bau/exact

      \ modulo
      bamod  baumod

      \ powers
      ba^pow-mod  ba^pow-mod-sec  ba^u-mod  ba^u  u^p>ba

      \ roots
      ba-nroot  ba-nroot-rem  ba-sqrt  ba-sqrt-rem

      \ number theoretics
      ba-next-prime  ba-gcd  bau-gcd  ba-gcd-st  ba-gcd-s
      ba-lcm  bau-lcm  ba-mod-invert  ba-remove
      fac>ba  ba-bin   bin>ba
      fib>ba  fib2>ba  luc>ba  luc2>ba 

      \ logic and bits
      ba-and  ba-or  ba-xor  ba-invert
      ba-bit-on  ba-bit-off  ba-bit-flip  ba-bit-test

  2.5 ARITHMETIC
      n>b  u>b  df>b
      b>n  b>u  b>df  b>df-exp
      
      \ elementary
      -b  |b|
      b+   b-   bu+   bu-  ub-  b*  bn*  bu*
      b*+  b*-  bu*+  bu*-
      b2^u*

      \ ceiled division
      bc/   bcrem   bc/rem
      bcu/  bcurem  bcu/rem  bc/urem  b2^uc/  b2^ucrem

      \ floored division
      bf/   bfrem   bf/rem
      bfu/  bfurem  bfu/rem  bf/urem  b2^uf/  b2^ufrem

      \ symmetric division
      bs/   bsrem   bs/rem
      bsu/  bsurem  bsu/rem  bs/urem  b2^us/  b2^usrem

      \ exact division
      b/exact  bu/exact

      \ modulo
      bmod  bumod

      \ divisibility and congruence
      bdivisible  bu-divisible  b2^u-divisible
      bcongruent  bu-congruent  b2^u-congruent

      \ powers
      b^pow-mod  b^pow-mod-sec  b^u-mod  b^u  u^p>b

      \ roots
      bnroot  bnroot-rem  bsqrt  bsqrt-rem
      bperfect-pow  bperfect-squ

      \ number theoretics
      bprob-prime  bnext-prime
      bgcd  bu-gcd>u  bu-gcd  bgcd-st  bgcd-s
      blcm  bu-lcm  bmod-invert
      bjacobi  blegendre  bkronecker
      bn-kronecker  bu-kronecker  nb-kronecker  ub-kronecker
      bremove  fac>b  bu-bin  bin>b  fib>b  fib2>b  luc>b  luc2>b

      \ logic and bits
      band  bor  bxor  binvert
      bbit-on  bbit-off  bbit-flip  bbit-test

  2.6 COMPARISONS
      bsgn
      b>      b=      b<      b>=      b<= 
      b0>     b0=     b0<     b0>=     b0<=
      bn>     bn=     bn<     bn>=     bn<=
      bu>     bu=     bu<     bu>=     bu<=
      bdf>    bdf=    bdf<    bdf>=    bdf<=
      b|>|    b|=|    b|<|    b|>=|    b|<=|
      bu|>|   bu|=|   bu|<|   bu|>=|   bu|<=| 
      bdf|>|  bdf|=|  bdf|<|  bdf|>=|  bdf|<=|

  3 RATIONALS

  3.1 RECORD SPACE
      DEFAULT-/QBUF  QSPACE  make-qspace  free-qrecs

  3.2 GARBAGE COLLECTION
      collect-qgarbage  qgarbage?
      qgc-off  qgc-on  qgc-lock@  qgc-lock!

  3.3 STACK
      (q:  qunused  qdepth  qs@  qs2@  qs3@  >qs  >qs-copy
      QVARIABLE  QCONSTANT  free-qconstants
      q"  q@  q!  qdrop  q2drop  q3drop  qdup  q2dup  qswap
      qnip  qover  qtuck  qpick  q.

  3.4 ACCUMULATOR ARITHMETIC
      q>qa  qa>q  0qa  nu>qa  uu>qa  df>qa  qa>df  qacanon  qa.

      \ elementary
      -qa  |qa|	 1/qa  qa+  qa-	 qa*  qa/  qa2^u*  qa2^u/

  3.5 ARITHMETIC
      nu>q  uu>q  df>q  q>df  qcanon

      \ elementary
      -q  |q|  1/q  q+  q-  q*  q/  q2^u*  q2^u/

  3.6 COMPARISONS
      qsgn
      q>      q=      q<      q>=      q<= 
      q0>     q0=     q0<     q0>=     q0<=
      qnu>    qnu=    qnu<    qnu>=    qnu<=
      quu>    quu=    quu<    quu>=    quu<=

  4 FLOATING POINT

  4.1 DEFAULT PRECISION
      set-default-mfprecision  get-default-mfprecision

  4.2 RECORD SPACE
      DEFAULT-/MFBUF  MFSPACE  make-mfspace  free-mfrecs
      fa-start-precision  fa-init-precision

  4.3 GARBAGE COLLECTION
      collect-mfgarbage  mfgarbage?
      mfgc-off  mfgc-on  mfgc-lock@  mfgc-lock!

  4.4 STACK
      (mf:  mfunused  mfdepth  mfs@  mfs2@  mfs3@  >mfs  >mfs-copy
      MFVARIABLE  MFCONSTANT  free-mfconstants
      mf"  mf@  mf!  mfdrop  mf2drop  mf3drop  mfdup  mf2dup  mfswap
      mfnip  mfover  mftuck  mfpick  mf.

  4.5 ACCUMULATOR ARITHMETIC
      mf>fa  fa>mf  0fa
      n>fa   u>fa   df>fa
      fa>n   fa>u   fa>df  fa>df-exp  fa.
      set-fa-precision  get-fa-precision

      \ elementary
      -fa  |fa|  fa+  fa-  fau+  fau-  ufa-  fa*  fau*  fa/  fau/  ufa/
      fasqrt  usqrt>fa  fa^u  fa2^u*  fa/2*u  farel-

      \ integer rounding
      fa-ceil fa-floor fa-trunc fa-integer

  4.6 ARITHMETIC
      n>mf  u>mf  df>mf
      mf>n  mf>u  mf>df  mf>df-exp

      \ elementary
      -mf  |mf|  mf+  mf-  mfu+  mfu- umf- mf*  mfu*  mf/  mfu/  umf/
      mfsqrt  usqrt>mf  mf^u  mf*2^u  mf/2^u  mfrel-

      \ integer rounding
      mf-ceil mf-floor mf-trunc mf-integer

  4.7 COMPARISONS
      mfsgn
      mf>    mf=    mf<    mf>=    mf<= 
      mf0>   mf0=   mf0<   mf0>=   mf0<=
      mfn>   mfn=   mfn<   mfn>=   mfn<=
      mfu>   mfu=   mfu<   mfu>=   mfu<=
      mfdf>  mfdf=  mfdf<  mfdf>=  mfdf<=

  5 MIXED STACK
      q>b    mf>b    b>q    fs>q    b>mf   q>mf
      qa>b   fa>b    ba>q   fa>q    ba>mf  qa>mf
      q>ba   mf>ba   b>qa   mf>qa   b>fa   q>fa
      qa>ba  fa>ba   ba>qa  fa>qa   ba>fa  qa>fa

      qnum>b   qden>b   qanum>b   qaden>b
      b>qnum   b>qden   b>qanum   b>qaden
      qnum>ba  qden>ba  qanum>ba  qaden>ba
      ba>qnum  ba>qden  ba>qanum  ba>qaden

  6 RANDOM NUMBERS

  6.1 STATE
      RANDSTATE:  free-randstate  free-randstates  copy-rand

  6.2 INITIALIZATION
      randinit-default  randinit-mt  randinit-lc-2^u  randinit-lc-bits

  6.3 SEEDING
      b>seed  u>seed

  6.4 GENERATION
      rand-#bits>u   rand[0,n-1]>u
      rand-#bits>ba  rand2-#bits>ba  rand-#bits>b   rand2-#bits>b
      rand-sbits>fa  rand2-limbs>fa  rand-sbits>mf  rand2-limbs>mf
)


\ *** 0 NOTATION AND TERMINOLOGY
---
There are three mathematical types, or objects, in libgmp,
integers, rationals (quotients), and floating-point numbers,
with C names mpz_t, mpq_t, and mpf_t.  Each of these corresponds
to a "record" in the terminology here, with a garbage-collected
space for each that includes a stack of record addresses.  There
is a fourth libgmp data type, gmp_randstate_t, whose references
do not have a separate stack.

b, q, mf:  integer big number, or just "big number", rational
  number with big-number numerator and denominator, or
  multiprecision floating-point (multifloat) number,
  corresponding to gmp mathemamtical objects.

brec, qrec, mfrec:  a b, q, or mf record, which manages
  information about the storage of the corresponding gmp data.

randstate:  a record that manages random number generation
  algorithm and data, corresponding to the C type gmp_randstate_t.

'b, 'q, 'f, 'rand:  the address of a brec, qrec, mfrec, or
  randstate.

dbrec, dqrec, dmfrec:  a brec, qrec, or mfrec in bspace, qspace,
  or mfspace, which means it is subject to garbage collection.
  The "d" stands for "dynamic".  A record outside of all of the
  three spaces is called "nondynamic" or "external".  Note that
  nondynamic records nevertheless manage memory dynamically
  allocated by libgmp.

bspace, qspace, mfspace:  a contiguous block of memory which can
  be allocated by the user (words provided below).  It contains
  a header and a buffer.  The buffer contains dynamic records
  building up from lower memory and a stack of dynamic and
  nondynamic record addresses building down from upper memory. 
  The names are also used for the values BSPACE, QSPACE, and
  MFSPACE, which hold the addresses of currently active spaces.

bstack, qstack, mfstack:  the stack in the corresponding space. 
  Although items on the stacks are addresses of records, their
  stack notation uses the style b, q, and f instead of 'b, 'q,
  and 'f.

accumulator:  a special, nondynamic brec, qrec, or mfrec used
  for sequential calculations and/or their results.  Currently
  there are BA, QA, and FA, plus two extra big-number
  accumulators, BY, and BZ, for secondary results such as
  remainders.  Accumulator arithmetic words have abbreviated
  stack effects, where what would be the leftmost stack input is
  instead in the accumulator, and what would be the leftmost
  stack output (usually the only one) instead remains in the
  accumulator.

[predicate]:  true/false corresponding to predicate.  E.g., [b1<b2].

zstr:  A C-style, null-terminated string, with address 'zstr.

"free":  Used in the names of words that release dynamic memory
  allocated by libgmp, and managed by a brec, qrec, mfrec, or
  randstate.  This is aside from garbage collection for dynamic
  records.  The libgmp function names use "clear" instead of
  "free".  The use of "free" aims to avoid the connotation of
  setting to zero, which is done instead by libgmp functions and
  words with "init" in their names.

u, n, +n, +p, etc.:  u is the standard Forth unsigned cell data
  type.  In this file, +n, +p, etc., are *never* the standard
  Forth signed, nonnegative data type +n.  Instead they are
  occasional names for u's, chosen to correspond to argument
  names in the libgmp documentation.  On the other hand, "n"
  without a "+" in stack comments *does* mean the standard Forth
  signed cell type.
---


\ *** 1 GENERAL SUPPORT
---
The values MZBUF and /MZBUF hold the address and size of a
buffer for transient, multiple precision string input and
output.  Strings in the buffer are measured (cell count field)
and null terminated, hence the MZ in the value names.  With some
restrictions (see the gmp documentation) input strings may
contain whitespace characters, including end of line characters.
---

1024 constant DEFAULT-/MZBUF
create DEFAULT-MZBUF DEFAULT-/MZBUF allot
DEFAULT-MZBUF value MZBUF    DEFAULT-/MZBUF value /MZBUF

: ?MZBUF ( u --  ) /MZBUF u> ABORT" ***MZBUF overflow" ;

: [MZ"]  \ compile:  ( "<lines><">" -- )
         \ run:      ( --  'zstr )
  ---
    A compile-time factor used by MZ", for string input across
    lines with null termination.
  ---
  state @ 0= IF -14 throw THEN
  postpone AHEAD [char] " upto-m,s drop null-m+,
  >r align postpone THEN r> postpone literal ;

: MZ" \ compile:    ( "<lines><">" -- )
      \ run:	    ( --  'zstr )
      \ interpret:  ( "<lines><">" -- 'zstr )
  STATE @ IF
    [MZ"]
  ELSE
    MZBUF /MZBUF m!"     0= ABORT" string buffer overflow"
    MZBUF /MZBUF null-m+ 0= ABORT" string buffer overflow"
    MZBUF cell+
  THEN ; immediate

[UNDEFINED] ZCOUNT [IF]
: ZCOUNT  ( 'zstr -- 'zstr len )
  dup 0 2>r
  BEGIN
    dup c@ 0<>
  WHILE
    1+ r> 1+ >r
  REPEAT drop 2r> ;
[THEN]

[UNDEFINED] ZSTRLEN [IF] : ZSTRLEN  ZCOUNT nip ; [THEN]

: wff  postpone 0= postpone 0= ; immediate  \ well formed flag

[UNDEFINED] -rot  [IF] : -rot rot rot ; [THEN]
[UNDEFINED] rdrop [IF] : rdrop postpone r> postpone drop ; immediate [THEN]
[UNDEFINED] (f:   [IF] : (f: postpone ( ; immediate [THEN]


\ *** 2 INTEGERS

\ *** 2.1 RECORD SPACE
---
MAKE-BSPACE uses ALLOCATE.  The resulting bspace can be freed
with FREE.  However, that should be done only after calling
FREE-BRECS.  See below.
---

: copy-brec ( 'b.src 'b.targ -- )
  ---
    Assume that target record memory has been reserved. 
    Initialize it and set it to the value of the source b. 
    Intended only for internal use by gcrstack.fs.
  ---
    swap mpz_init_set ;

0 value BSPACE

: make-bspace  ( /buf -- )
  ---
    This word writes over the value BSPACE.  Be sure that it's
    either saved or no longer needed.  The space is reserved
    with ALLOCATE, and can be released with BSPACE FREE. 
    However, that should be done only after executing
    FREE-BRECS.  See below.
  ---
  ['] mpz_clear ['] copy-brec rot /MPZ swap
  make-gspace to bspace ;

/MPZ 2 cells + 4000 * CONSTANT DEFAULT-/BBUF
DEFAULT-/BBUF make-bspace

: free-brecs ( -- )
  ---
    Free any unreleased records in bspace, zero any bvariables
    pointing to them, and clear bspace, including the bstack.
  ---
  bspace free-grecs ;

\ *** 2.2 GARBAGE COLLECTION

: bgarbage?	   ( -- flag ) bspace ggarbage? ;
: collect-bgarbage ( -- flag ) bspace collect-ggarbage ;
: bgc-off	   ( -- )      bspace ggc-off ;
: bgc-on	   ( -- )      bspace ggc-on ;
: bgc-lock@	   ( -- flag ) bspace ggc-lock@ ;
: bgc-lock!	   ( flag -- ) bspace ggc-lock! ;

\ *** 2.3 STACK

PFE-HOST    [IF]
  synonym (b: (
  synonym BVARIABLE GVARIABLE
[THEN]
GFORTH-HOST [IF]
  ' (         alias (b: immediate
  ' GVARIABLE alias BVARIABLE
[THEN]
IFORTH-HOST [IF]
  : (b:  postpone ( ; immediate
  : BVARIABLE  GVARIABLE ;
[THEN]

: bunused ( -- u )       bspace gunused ;
: bdepth  ( -- u )	 bspace gdepth ;
: (>bs)   ( 'b -- b: b ) bspace (>gs) ;
: (bdrop) (b: b -- )	 bspace (gdrop) ;

: bs@	   (b: b -- b s: bra )			     bspace gs@ ;
: bs2@	   (b: b1 b2 -- b1 b2 s: 'b1 'b2 )	     bspace gs2@ ;
: bs3@	   (b: b1 b2 b3 -- b1 b2 b3 s: 'b1 'b2 'b3 ) bspace gs3@ ;
: >bs	   ( 'b -- b: b )			     bspace >gs ;
: >bs-copy ( 'b -- b: b )			     bspace >gs-copy ;

: num-xt>bs  ( <num> xt -- b: num )
  ---
    Use xt to transfer <num>, a number on the data,
    floating-point, q, or mf stack, or a bnumber string in the
    input stream, into a dbrec, and push it onto the bstack. 
    The stack effect for xt is ( <num> 'b -- ).
  ---
  >r 
  bspace >/grec @ cell+ bspace ?GROOM
  bspace new-grec ( 'b) >r
  r@ mpz_init
  2r@ swap EXECUTE
  r@ bspace (gpush)
  bspace >gsp @ r> cell- !  \ link brec to bstack
  rdrop ;

\ private
: b">'b  ( "<chars><">" 'b -- )  \ conversion defends on BASE
  dup postpone mz" ( 'b 'zstr) base @ mpz_set_str
  IF ( 'b) dup mpz_clear
           cell- bspace >gbrkp !    \ restore bspace state 
     true ABORT" ***invalid gmp integer string" THEN
  ( 'b) drop ;

: b" ( "<chars><">" -- b: b ) ['] b">'b num-xt>bs ;

0 [IF]
: BCONSTANT ( "name" b: b -- )
  bspace gs>extern ( ra) create ,
DOES>  (b: -- b.ext )  @ (>bs) ;
[THEN]

0 value LAST-BCONSTANTP
: BCONSTANT ( "name" b: b -- )
  bspace gs>extern ( ra) >r create here r> ,
  last-bconstantp , to last-bconstantp
DOES>  (b: -- b.ext )  @ (>bs) ;

: free-bconstants ( -- )
  last-bconstantp
  BEGIN  dup 
  WHILE  dup @ mpz_clear cell+ @
  REPEAT drop ; 

: b,'b	 (b: b -- s: 'b )	    bspace g,ra ;
: b@	 ( bva -- b: b )	    bspace g@ ;
: b!	 (b: b s: bva -- )	    bspace g! ;
: bdrop  (b: b -- )		    bspace gdrop ; 
: b2drop (b: b1 b2 -- )		    bspace g2drop ;
: b3drop (b: b1 b2 b3 -- )	    bspace g3drop ;
: bdup   (b: b -- b b )		    bspace gdup ;
: b2dup  (b: b1 b2 -- b1 b2 b1 b2 ) bspace g2dup ;
: bswap  (b: b1 b2 -- b2 b1 )	    bspace gswap ;
: bnip   (b: b1 b2 -- b2 )	    bspace gnip ;
: bover  (b: b1 b2 -- b1 b2 b1 )    bspace gover ;
: btuck  (b: b1 b2 -- b2 b1 b2 )    bspace gtuck ;
: bpick  ( u b: bu ... b0 -- bu ... b0 bu ) bspace gpick ;

: b.  (b: b -- )
  bs@ base @
  mpz_sizeinbase [ 1 cells 2 chars + ] literal + ?MZBUF 
  MZBUF cell+ base @ bs@ mpz_get_str bdrop
  zcount dup MZBUF ! type space ;

\ *** 2.4 ACCUMULATOR ARITHMETIC
---
An accumulator can be used for a sequence of computations to
minimize bstack transfers, but especially to optimize the
initialization of brecords and the release of the memory they
manage.

Keep in mind that computation words that input only from and
output only to the bstack overwrite not only the BA accumulator,
but also BY, when a bremainder is returned, and BY and BZ, when
gcd coefficients are returned.
---

CREATE ba /MPZ allot ba mpz_init
CREATE by /MPZ allot by mpz_init
CREATE bz /MPZ allot by mpz_init

: 0ba   ( -- )	   ba mpz_clear ba mpz_init ;
: 0by   ( -- )	   by mpz_clear by mpz_init ;
: 0bz   ( -- )	   bz mpz_clear bz mpz_init ;
: b>ba  (b: b -- ) ba dup mpz_clear bs@ mpz_init_set bdrop ;
: n>ba  ( n -- )   ba dup mpz_clear swap mpz_init_set_si ;
: u>ba  ( u -- )   ba dup mpz_clear swap mpz_init_set_ui ;
: df>ba ( n -- )   ba dup mpz_clear mpz_init_set_d ;

\ Assume that BA, BY, and BZ have been initialized.

: ba>b (b: -- ba ) ba >bs-copy ;
: by>b (b: -- y )  by >bs-copy ;
: bz>b (b: -- z )  bz >bs-copy ;

: ba>n	    ( -- n )	      ba mpz_get_si ;
: ba>u	    ( -- u )	      ba mpz_get_ui ;
: ba>df	    (f: -- df )	      ba mpz_get_d ;
: ba>df-exp ( 'exp -- f: df ) ba mpz_get_d_2exp ;

: ba. ( -- ) ba>b b. ;

\ elementary
: -ba	   ( -- )	   ba dup mpz_neg ;
: |ba|	   ( -- )	   ba dup mpz_abs ;
: ba+	   (b: b -- )	   ba dup bs@ mpz_add bdrop ;
: ba-	   (b: b -- )	   ba dup bs@ mpz_sub bdrop ;
: bau+	   ( u -- )	   ba dup  rot mpz_add_ui ;
: bau-	   ( u -- )	   ba dup  rot mpz_sub_ui ;
: uba-     ( u -- )	   ba dup -rot mpz_ui_sub ;
: ba*	   (b: b -- )	   ba dup bs@ mpz_mul bdrop ;
: ban*	   ( n -- )	   ba dup rot mpz_mul_si ;
: bau*	   ( u -- )	   ba dup rot mpz_mul_ui ;
: ba*+ 	   (b: b1 b2 -- )  ba bs2@ mpz_addmul b2drop ;
: bau*+	   (b: b s: u -- ) ba bs@ rot mpz_addmul_ui bdrop ;
: ba*-	   (b: b1 b2 -- )  ba bs2@ mpz_submul b2drop ;
: bau*-	   (b: b s: u -- ) ba bs@ rot mpz_submul_ui bdrop ;
: ba2^u* ( #bits -- )	   >r ba dup r> mpz_mul_2exp ;

\ ceiled division
: bac/      (b: den -- )   ba dup bs@ mpz_cdiv_q bdrop ;
: bacrem    (b: den -- )   ba dup bs@ mpz_cdiv_r bdrop ;
: bac/rem   (b: den -- )   ba by ba bs@ mpz_cdiv_qr bdrop ;
: bauc/     ( u -- |rem| ) ba dup rot mpz_cdiv_q_ui ;
: baucrem   ( u -- |rem| ) ba dup rot mpz_cdiv_r_ui ;
: bauc/rem  ( u -- |rem| ) >r ba by ba r> mpz_cdiv_qr_ui ;
: bacurem   ( u -- |rem| ) ba swap mpz_cdiv_ui ;  \ unfortunate name
: ba2^uc/   ( pow -- )	   ba dup rot mpz_cdiv_q_2exp ;
: ba2^ucrem ( pow -- )	   ba dup rot mpz_cdiv_r_2exp ;

\ floored division
: baf/      (b: den -- )   ba dup bs@ mpz_fdiv_q bdrop ;
: bafrem    (b: den -- )   ba dup bs@ mpz_fdiv_r bdrop ;
: baf/rem   (b: den -- )   ba by ba bs@ mpz_fdiv_qr bdrop ;
: bauf/     ( u -- |rem| ) ba dup rot mpz_fdiv_q_ui ;
: baufrem   ( u -- |rem| ) ba dup rot mpz_fdiv_r_ui ;
: bauf/rem  ( u -- |rem| ) >r ba by ba r> mpz_fdiv_qr_ui ;
: bafurem   ( u -- |rem| ) ba swap mpz_fdiv_ui ;  \ unfortunate name
: ba2^uf/   ( pow -- )	   ba dup rot mpz_fdiv_q_2exp ;
: ba2^ufrem ( pows -- )	   ba dup rot mpz_fdiv_r_2exp ;

\ symmetric division
: bas/      (b: den -- )   ba dup bs@ mpz_tdiv_q bdrop ;
: basrem    (b: den -- )   ba dup bs@ mpz_tdiv_r bdrop ;
: bas/rem   (b: den -- )   ba by ba bs@ mpz_tdiv_qr bdrop ;
: baus/     ( u -- |rem| ) ba dup rot mpz_tdiv_q_ui ;
: bausrem   ( u -- |rem| ) ba dup rot mpz_tdiv_r_ui ;
: baus/rem  ( u -- |rem| ) >r ba by ba r> mpz_tdiv_qr_ui ;
: basurem   ( u -- |rem| ) ba swap mpz_tdiv_ui ;  \ unfortunate name
: ba2^us/   ( pow -- )	   ba dup rot mpz_tdiv_q_2exp ;
: ba2^usrem ( pow -- )	   ba dup rot mpz_tdiv_r_2exp ;

\ exact division
: ba/exact  (b: den -- ) ba dup bs@ mpz_divexact bdrop ;
: bau/exact ( u -- )	 ba dup rot mpz_divexact_ui ;

\ modulo
: bamod	 (b: den -- )      ba dup bs@ mpz_mod bdrop ;
: baumod ( u -- ba.mod.u ) ba dup rot mpz_mod_ui ;

\ powers
: ba^pow-mod     (b: pow mod -- ) ba dup bs2@ mpz_powm b2drop ; 
: ba^pow-mod-sec (b: pow mod -- ) ba dup bs2@ mpz_powm_sec b2drop ;
: ba^u-mod	 ( u b: mod -- )  ba dup rot bs@ mpz_powm_ui bdrop ;
: ba^u		 ( u -- )	  ba dup rot mpz_pow_ui ;
: u^p>ba	 ( u +p -- )	  0ba ba -rot mpz_ui_pow_ui ;

\ roots
: ba-nroot     ( u -- flag ) ba dup rot mpz_root wff ;
: ba-nroot-rem ( u -- )	     >r ba by ba r> mpz_rootrem ;
: ba-sqrt      ( -- )	     ba dup mpz_sqrt ;
: ba-sqrt-rem  ( -- )	     ba by ba mpz_sqrtrem ; 

\ number theoretics
: ba-next-prime	( -- )	       ba dup mpz_nextprime ;
: ba-gcd	(b: b -- )     ba dup bs@ mpz_gcd bdrop ;
: bau-gcd	( u -- gcd|0 ) ba dup rot mpz_gcd_ui ; 
: ba-gcd-st	(b: b -- )     ba 0by 0bz by bz ba bs@ mpz_gcdext bdrop ;
: ba-gcd-s	(b: b -- )     ba 0by by 0 ba bs@ mpz_gcdext bdrop ;
: ba-lcm	(b: b -- )     ba dup bs@ mpz_lcm bdrop ;
: bau-lcm	( u -- )       ba dup rot mpz_lcm_ui ;
: ba-mod-invert (b: mod -- s: flag ) ba dup bs@ mpz_invert bdrop wff ;
: ba-remove	(b: b factor -- s: #removed ) ba dup bs@ mpz_remove bdrop ;
: fac>ba	( +n -- )      0ba ba swap mpz_fac_ui ;
: ba-bin	( +k -- )      ba dup rot mpz_bin_ui ;
: bin>ba	( +n +k -- )   0ba ba -rot mpz_bin_uiui ;
: fib>ba	( +n -- )      0ba ba swap mpz_fib_ui ;
: fib2>ba	( +n -- )      0ba ba 0by by rot mpz_fib2_ui ;
: luc>ba	( +n -- )      0ba ba swap mpz_lucnum_ui ;
: luc2>ba	( +n -- )      0ba ba 0by by rot mpz_lucnum2_ui ;

\ logic and bits
: ba-and      (b: b -- )       ba dup bs@ mpz_and bdrop ;
: ba-or	      (b: b -- )       ba dup bs@ mpz_ior bdrop ;
: ba-xor      (b: b -- )       ba dup bs@ mpz_xor bdrop ;
: ba-invert   ( -- )	       ba dup mpz_com ;
: ba-bit-on   ( bit# -- )      ba swap mpz_setbit ;
: ba-bit-off  ( bit# -- )      ba swap mpz_clrbit ;
: ba-bit-flip ( bit# -- )      ba swap mpz_combit ;
: ba-bit-test ( bit# -- flag ) ba swap mpz_tstbit wff ;

\ *** 2.5 ARITHMETIC

\ private
: n>'b  ( n 'b -- )	  swap mpz_set_si ;
: u>'b  ( u 'b -- )	  swap mpz_set_ui ;
: df>'b (f: df s: 'b -- ) mpz_set_d ;

: n>b  ( n -- b: n )	   ['] n>'b num-xt>bs ;
: u>b  ( u -- b: u )	   ['] u>'b num-xt>bs ;
: df>b (f: df -- b: [df] ) ['] df>'b num-xt>bs ;

: b>n	    (b: b -- s: n )	   bs@ mpz_get_si  bdrop ; 
: b>u	    (b: b -- s: |b| )	   bs@ mpz_get_ui bdrop ;
: b>df	    (b: b -- f: b )	   bs@ mpz_get_d bdrop ;
: b>df-exp (b: b s: 'exp -- f: b ) bs@ mpz_get_d_2exp bdrop ;

\ elementary
: -b	(b: b -- -b )			 b>ba -ba ba>b ;
: |b|	(b: b -- |b|)			 b>ba |ba| ba>b ;
: b+	(b: b1 b2 -- b1+b2 )		 b>ba ba+ ba>b ;
: b-	(b: b1 b2 -- b1-b2 )		 b>ba ba- -ba ba>b ;
: bu+	(b: b s: u -- b: b+u )		 b>ba bau+ ba>b ;
: bu-	(b: b s: u -- b: b-u )		 b>ba bau- ba>b ;
: ub-	(b: b s: u -- b: u-b )		 b>ba uba- ba>b ;
: b*	(b: b1 b2 -- b1*b2 )		 b>ba ba*  ba>b ;
: bn*	(b: b s: n -- b: b*n )		 b>ba ban* ba>b ;
: bu*	(b: b s: u -- b: b*u )		 b>ba bau* ba>b ;
: b*+	(b: b b1 b2 -- b+b1*b2 )	 2 bpick b>ba ba*+ bdrop ba>b ;
: b*-	(b: b b1 b2 -- b+b1*b2 )	 2 bpick b>ba ba*- bdrop ba>b ;
: bu*+	(b: b b1 s:u -- b+b1*u )	 bswap b>ba bau*+ ba>b ;
: bu*-	(b: b b1 s:u -- b-b1*u )	 bswap b>ba bau*- ba>b ;
: b2^u* (b: b s: #bits -- b: b*2^#bits ) b>ba ba2^u* ba>b ;

\ ceiled division
: bc/      (b: num den -- quo )		      bswap b>ba bac/ ba>b ;
: bcrem    (b: num den -- rem )		      bswap b>ba bacrem ba>b ;
: bc/rem   (b: num den -- quo rem )	      bswap b>ba bac/rem ba>b by>b ; 
: buc/	   (b: num s: u -- |rem| b: quo )     b>ba bauc/ ba>b ;
: bucrem   (b: num s: u -- |rem| b: rem )     b>ba baucrem ba>b ;
: buc/rem  (b: num s: u -- |rem |b: quo rem ) b>ba bauc/rem ba>b by>b ;
: bcurem   (b: num s: u -- |rem| )	      b>ba bacurem ;
: b2^uc/   (b: num s: #bits -- b: quo )	      b>ba ba2^uc/ ba>b ;
: b2^ucrem (b: num s: #bits -- b: rem )	      b>ba ba2^ucrem ba>b ;

\ floored division
: bf/      (b: num den -- quo )		      bswap b>ba baf/ ba>b ;
: bfrem    (b: num den -- rem )		      bswap b>ba bafrem ba>b ;
: bf/rem   (b: num den -- quo rem )	      bswap b>ba baf/rem ba>b by>b ; 
: buf/	   (b: num s: u -- |rem| b: quo )     b>ba bauf/ ba>b ;
: bufrem   (b: num s: u -- |rem| b: rem )     b>ba baufrem ba>b ;
: buf/rem  (b: num s: u -- |rem |b: quo rem ) b>ba bauf/rem ba>b by>b ;
: bfurem   (b: num s: u -- |rem| )	      b>ba bafurem ;
: b2^uf/   (b: num s: #bits -- b: quo )	      b>ba ba2^uf/ ba>b ;
: b2^ufrem (b: num s: #bits -- b: rem )	      b>ba ba2^ufrem ba>b ;

\ symmetric division
: bs/      (b: num den -- quo )		      bswap b>ba bas/ ba>b ;
: bsrem    (b: num den -- rem )		      bswap b>ba basrem ba>b ;
: bs/rem   (b: num den -- quo rem )	      bswap b>ba bas/rem ba>b by>b ; 
: bus/	   (b: num s: u -- |rem| b: quo )     b>ba baus/ ba>b ;
: busrem   (b: num s: u -- |rem| b: rem )     b>ba bausrem ba>b ;
: bus/rem  (b: num s: u -- |rem |b: quo rem ) b>ba baus/rem ba>b by>b ;
: bsurem   (b: num s: u -- |rem| )	      b>ba basurem ;
: b2^us/   (b: num s: #bits -- b: quo )	      b>ba ba2^us/ ba>b ;
: b2^usrem (b: num s: #bits -- b: rem )	      b>ba ba2^usrem ba>b ;

\ exact division
: b/exact  (b: num den -- num/den )	  bswap b>ba ba/exact ba>b ;
: bu/exact (b: num s: den -- b: num/den ) b>ba bau/exact ba>b ;

\ modulo
: bmod  (b: num den -- mod )	bswap b>ba bamod ba>b ;
: bumod (b: num s: den -- mod )	b>ba baumod ba>b ;

\ divisibility and congruence
: bdivisible	 (b: num den -- s: flag )   bs2@ mpz_divisible_p wff b2drop ;
: bu-divisible	 (b: num s: den -- flag )   bs@ swap mpz_divisible_ui_p wff bdrop ;
: b2^u-divisible (b: num s: pow -- flag )   bs@ swap mpz_divisible_2exp_p wff bdrop ;
: bcongruent	 (b: num c den -- s: flag ) bs3@ mpz_congruent_p wff b3drop ;
: bu-congruent	 (b: num s: c den -- flag ) bs@ -rot mpz_congruent_ui_p wff bdrop ;
: b2^u-congruent (b: num c s: pow -- flag ) bs2@ rot mpz_congruent_2exp_p wff b2drop ;

\ powers
: b^pow-mod	(b: b pow mod -- b' )	  2 bpick b>ba ba^pow-mod bdrop ba>b ;
: b^pow-mod-sec (b: b pow mod -- b' )	  2 bpick b>ba ba^pow-mod-sec bdrop ba>b ;
: b^u-mod	(b: b s: u b: mod -- b' ) bswap b>ba ba^u-mod ba>b ;
: b^u		(b: b s: u -- b: b^u )	  b>ba ba^u ba>b ;
: u^p>b		( u +p -- b: u^p )	  u^p>ba ba>b ;

\ roots
: bnroot	(b: b s: u -- flag b: root ) b>ba ba-nroot ba>b ;
: bnroot-rem	(b: b s: u -- b: root rem )  b>ba ba-nroot-rem ba>b by>b ;
: bsqrt		(b: b -- root )		     b>ba ba-sqrt ba>b ;
: bsqrt-rem	(b: b -- root rem )	     b>ba ba-sqrt-rem ba>b by>b ; 
: bperfect-pow	(b: b -- flag )		     bs@ mpz_perfect_power_p wff bdrop ;
: bperfect-squ	(b: b -- flag )		     bs@ mpz_perfect_square_p wff bdrop ;

\ number theoretics
: bprob-prime  (b: b s: #reps -- 2|1|0 )    bs@ swap mpz_probab_prime_p bdrop ;
: bnext-prime  (b: b -- prime )		    b>ba ba-next-prime ba>b ;
: bgcd	       (b: b1 b2 -- gcd )	    b>ba ba-gcd ba>b ;
: bu-gcd>u     (b: b s: u -- gcd|0 )	    0 bs@ rot mpz_gcd_ui bdrop ;
: bu-gcd  (b: b s: u -- b: gcd|b s: gcd|0 ) b>ba bau-gcd ba>b ;
: bgcd-st      (b: b1 b2 -- s t gcd )	    bswap b>ba ba-gcd-st by>b bz>b ba>b ; 
: bgcd-s       (b: b1 b2 -- s gcd )	    bswap b>ba ba-gcd-s by>b ba>b ; 
: blcm	       (b: b1 b2 -- lcm )	    b>ba ba-lcm ba>b ;
: bu-lcm       (b: b s: u -- b: lcm )	    b>ba bau-lcm ba>b ;
: bmod-invert  (b: b mod -- inv|0 s: flag ) bswap b>ba ba-mod-invert ba>b ;
: bjacobi      (b: a b -- s: jac )	    bs2@ mpz_jacobi b2drop ;
: blegendre    (b: a b -- s: leg )	    bs2@ mpz_legendre b2drop ;
: bkronecker   (b: a b -- s: kro )	    bs2@ mpz_kronecker b2drop ;
: bn-kronecker (b: a s: n -- kro )	    bs@ swap mpz_kronecker_si bdrop ;
: bu-kronecker (b: a s: u -- kro )	    bs@ swap mpz_kronecker_ui bdrop ;
: nb-kronecker (b: b s: n -- kro )	    bs@ mpz_si_kronecker bdrop ;
: ub-kronecker (b: b s: u -- kro )	    bs@ mpz_ui_kronecker bdrop ;
: bremove  (b: b factor -- b' s: #removed ) bswap b>ba ba-remove ba>b ;
: fac>b	       ( +n -- b: n! )		    fac>ba ba>b ; 
: bu-bin       (b: b s: +k -- b: bin )	    b>ba ba-bin ba>b ;
: bin>b	       ( +n +k -- b: bin )	    bin>ba ba>b ;
: fib>b	       ( +n -- b: fib )	            fib>ba ba>b ;
: fib2>b       ( +n -- b: fib_[n-1] fib_n ) fib2>ba by>b ba>b ;
: luc>b	       ( +n -- b: luc )		    luc>ba ba>b ;
: luc2>b       ( +n -- b: luc_[n-1] luc_n ) luc2>ba by>b ba>b ;

\ logic and bits
: band	     (b: b1 b2-- )	       b>ba ba-and ba>b ;
: bor	     (b: b1 b2 -- )	       b>ba ba-or ba>b ;
: bxor	     (b: b1 b2 -- )	       b>ba  ba-xor ba>b ;
: binvert    (b: b -- ~b)	       b>ba ba-invert ba>b ;
: bpop-count (b: b -- s: #bits )       bs@ mpz_popcount bdrop ;
: bham-dist  (b: b1 b2 -- s: #bits )   bs2@ mpz_hamdist b2drop ; 
: bscan0     (b: b s: start# -- bit# ) bs@ swap mpz_scan0 bdrop ;
: bscan1     (b: b s: start# -- bit# ) bs@ swap mpz_scan1 bdrop ;
: bbit-on    (b: b s: bit# -- b: b' )  b>ba ba-bit-on ba>b ;
: bbit-off   (b: b s: bit# -- b: b' )  b>ba ba-bit-off ba>b ;
: bbit-flip  (b: b s: bit# -- b: b' )  b>ba ba-bit-flip ba>b ;
: bbit-test  (b: b s: bit# -- flag )   b>ba ba-bit-test ;

\ *** 2.6 COMPARISONS

: bsgn (b: b -- s: sgn[b] ) bs@ mpz_sgn bdrop ;

: b>  (b: b1 b2 -- s: [b1>b2] )  bs2@ mpz_cmp 0>  b2drop ;
: b<  (b: b1 b2 -- s: [b1<b2] )  bs2@ mpz_cmp 0<  b2drop ;
: b=  (b: b1 b2 -- s: [b1=b2] )  bs2@ mpz_cmp 0=  b2drop ;
: b<= (b: b1 b2 -- s: [b1<=b2] ) bs2@ mpz_cmp 0<= b2drop ;
: b>= (b: b1 b2 -- s: [b1>=b2] ) bs2@ mpz_cmp 0>= b2drop ;

: b0>  (b: b -- s: [b>0] )  bs@ 0 mpz_cmp_si 0> bdrop ;
: b0=  (b: b -- s: [b=0] )  bs@ 0 mpz_cmp_si 0= bdrop ;
: b0<  (b: b -- s: [b=0] )  bs@ 0 mpz_cmp_si 0< bdrop ;
: b0>= (b: b -- s: [b>=0] ) bs@ 0 mpz_cmp_si dup 0> swap 0= or bdrop ;
: b0<= (b: b -- s: [b<=0] ) bs@ 0 mpz_cmp_si dup 0< swap 0= or bdrop ;

: bn>  (b: b s: n -- [b>n] )  bs@ swap mpz_cmp_si 0> bdrop ;
: bn=  (b: b s: n -- [b=n] )  bs@ swap mpz_cmp_si 0= bdrop ;
: bn<  (b: b s: n -- [b<n] )  bs@ swap mpz_cmp_si 0< bdrop ;
: bn>= (b: b s: n -- [b>=n] ) bs@ swap mpz_cmp_si dup 0> swap 0= or bdrop ;
: bn<= (b: b s: n -- [b<=n] ) bs@ swap mpz_cmp_si dup 0< swap 0= or bdrop ;

: bu>  (b: b s: u -- [b>u] )  bs@ swap mpz_cmp_ui 0> bdrop ;
: bu=  (b: b s: u -- [b=u] )  bs@ swap mpz_cmp_ui 0= bdrop ;
: bu<  (b: b s: u -- [b<u] )  bs@ swap mpz_cmp_ui 0< bdrop ;
: bu>= (b: b s: u -- [b>=u] ) bs@ swap mpz_cmp_ui dup 0> swap 0= or bdrop ;
: bu<= (b: b s: u -- [b<=u] ) bs@ swap mpz_cmp_ui dup 0< swap 0= or bdrop ;

: bdf>  (b: b f: df -- s: [b>df] )  bs@ mpz_cmp_d 0> bdrop ;
: bdf=  (b: b f: df -- s: [b=df] )  bs@ mpz_cmp_d 0= bdrop ;
: bdf<  (b: b f: df -- s: [b<df] )  bs@ mpz_cmp_d 0< bdrop ;
: bdf>= (b: b f: df -- s: [b>=df] ) bs@ mpz_cmp_d dup 0> swap 0= or bdrop ;
: bdf<= (b: b f: df -- s: [b<=df] ) bs@ mpz_cmp_d dup 0< swap 0= or bdrop ;

: b|>|  (b: b1 b2 -- s: [|b1|>|b2|] )  bs2@ mpz_cmpabs 0>  b2drop ;
: b|=|  (b: b1 b2 -- s: [|b1|=|b2|] )  bs2@ mpz_cmpabs 0=  b2drop ;
: b|<|  (b: b1 b2 -- s: [|b1|<|b2|] )  bs2@ mpz_cmpabs 0<  b2drop ;
: b|>=| (b: b1 b2 -- s: [|b1|>=|b2|] ) bs2@ mpz_cmpabs 0>= b2drop ;
: b|<=| (b: b1 b2 -- s: [|b1|<=|b2|] ) bs2@ mpz_cmpabs 0<= b2drop ;

: bdf|>|  (b: b f: df -- s: [|b|>|df|] )  bs@ mpz_cmpabs_d 0> bdrop ;
: bdf|=|  (b: b f: df -- s: [|b|=|df|] )  bs@ mpz_cmpabs_d 0= bdrop ;
: bdf|<|  (b: b f: df -- s: [|b|<|df|] )  bs@ mpz_cmpabs_d 0< bdrop ;
: bdf|>=| (b: b f: df -- s: [|b|>=|df|] ) bs@ mpz_cmpabs_d dup 0> swap 0= or bdrop ;
: bdf|<=| (b: b f: df -- s: [b|<|=|df|] ) bs@ mpz_cmpabs_d dup 0< swap 0= or bdrop ;

: bu|>|  (b: b s: u -- [|b|>u] )  bs@ swap mpz_cmpabs_ui 0> bdrop ;
: bu|=|  (b: b s: u -- [|b|=u] )  bs@ swap mpz_cmpabs_ui 0= bdrop ;
: bu|<|  (b: b s: u -- [|b|<u] )  bs@ swap mpz_cmpabs_ui 0< bdrop ;
: bu|>=| (b: b s: u -- [|b|>=u] ) bs@ swap mpz_cmpabs_ui dup 0> swap 0= or bdrop ;
: bu|<=| (b: b s: u -- [|b|<=u] ) bs@ swap mpz_cmpabs_ui dup 0< swap 0= or bdrop ;


\ *** 3 RATIONALS

\ *** 3.1 RECORD SPACE

: copy-qrec ( 'q.src 'q.targ -- )
  ---
    Assume that memory for the target q has been reserved.
    Initialize and set it to the same value as the source. 
    Intended only for internal use by gcrstack.fs.
  ---
  dup mpq_init swap mpq_set ;

0 value QSPACE

: make-qspace  ( /buf -- )
  ---
    This word writes over the value QSPACE.  Be sure that it's
    either saved or no longer needed.  The space is reserved
    with ALLOCATE, and can be released with QSPACE FREE. 
    However, that should be done only after executing
    FREE-QRECS.  See below.
  ---
  ['] mpq_clear ['] copy-qrec rot /MPQ swap
  make-gspace to qspace ;

/MPQ 2 cells + 4000 * CONSTANT DEFAULT-/QBUF
DEFAULT-/QBUF make-qspace

: free-qrecs ( -- )
  ---
    Free any unreleased records in qspace, zero any qvariables
    pointing to them, and clear qspace, including the qstack.
  ---
  qspace free-grecs ;

\ *** 3.2 GARBAGE COLLECTION

: qgarbage?	   ( -- flag ) qspace ggarbage? ;
: collect-qgarbage ( -- flag ) qspace collect-ggarbage ;
: qgc-off	   ( -- )      qspace ggc-off ;
: qgc-on	   ( -- )      qspace ggc-on ;
: qgc-lock@	   ( -- flag ) qspace ggc-lock@ ;
: qgc-lock!	   ( flag -- ) qspace ggc-lock! ;

\ *** 3.3 STACK

PFE-HOST    [IF]
  synonym (q: (
  synonym QVARIABLE GVARIABLE
[THEN]
GFORTH-HOST [IF]
  ' (         alias (q: immediate
  ' GVARIABLE alias QVARIABLE
[THEN]
IFORTH-HOST [IF]
  : (q:  postpone ( ; immediate
  : QVARIABLE  GVARIABLE ;
[THEN]

: qunused ( -- u )	 qspace gunused ;
: qdepth  ( -- u )	 qspace gdepth ;
: (>qs)	  ( 'q -- q: q ) qspace (>gs) ;
: (qdrop) (q: q -- )	 qspace (gdrop) ;

: qs@	   (q: q -- q s: qra )			     qspace gs@ ;
: qs2@	   (q: q1 q2 -- q1 q2 s: 'q1 'q2 )	     qspace gs2@ ;
: qs3@	   (q: q1 q2 q3 -- q1 q2 q3 s: 'q1 'q2 'q3 ) qspace gs3@ ;
: >qs	   ( 'q -- q: q )			     qspace >gs ;
: >qs-copy ( 'q -- q: q )			     qspace >gs-copy ;

: num-xt>qs  ( <num> xt -- q: num )
  ---
    Use xt to transfer <num>, a number on the data, b, or mf
    stack, or a q string in the input stream, into a dqrec, and
    push it onto the qstack.  The stack effect for xt is
    ( <num> 'q -- ).
  ---
  >r 
  qspace >/grec @ cell+ qspace ?GROOM
  qspace new-grec ( 'q) >r
  r@ mpq_init
  2r@ swap EXECUTE
  r@ qspace (gpush)
  qspace >gsp @ r> cell- !  \ link qrec to qstack
  rdrop ;

\ private
: q">'q  ( "<chars><">" 'q -- )     \ conversion depends on BASE
  dup postpone mz" ( 'q 'zstr) base @ mpq_set_str
  IF ( 'q) dup mpq_clear
           cell- qspace >gbrkp !    \ restore qspace state 
     true ABORT" ***invalid gmp rational string" THEN
  ( 'q) drop ;

: q" ( "<chars><">" -- q: q ) ['] q">'q num-xt>qs ;

0 [IF]
: QCONSTANT ( "name" q: q -- )
  qspace gs>extern ( ra) create ,
DOES>  (q: -- q.ext )  @ (>qs) ;
[THEN]

0 value LAST-QCONSTANTP
: QCONSTANT ( "name" q: q -- )
  qspace gs>extern ( ra) >r create here r> ,
  last-qconstantp , to last-qconstantp
DOES>  (q: -- q.ext )  @ (>qs) ;

: free-qconstants ( -- )
  last-qconstantp
  BEGIN  dup 
  WHILE  dup @ mpq_clear cell+ @
  REPEAT drop ; 

: q,'q	 (q: q -- s: 'q )		    qspace g,ra ;
: q@	 ( qva -- q: q )		    qspace g@ ;
: q!	 (q: q s: qva -- )		    qspace g! ;
: qdrop  (q: q -- )			    qspace gdrop ; 
: q2drop (q: q1 q2 -- )			    qspace g2drop ;
: q3drop (q: q1 q2 q3 -- )		    qspace g3drop ;
: qdup   (q: q -- q q )			    qspace gdup ;
: q2dup  (q: q1 q2 -- q1 q2 q1 q2 )	    qspace g2dup ;
: qswap  (q: q1 q2 -- q2 q1 )		    qspace gswap ;
: qnip   (q: q1 q2 -- q2 )		    qspace gnip ;
: qover  (q: q1 q2 -- q1 q2 q1 )	    qspace gover ;
: qtuck  (q: q1 q2 -- q2 q1 q2 )	    qspace gtuck ;
: qpick  ( u q: qu ... q0 -- qu ... q0 qu ) qspace gpick ;

: q.  (q: q -- )
  qs@ mpq_numref base @ mpz_sizeinbase
  qs@ mpq_denref base @ mpz_sizeinbase + cell+ 3 chars + ?MZBUF 
  MZBUF cell+ base @ qs@ mpq_get_str qdrop
  zcount dup MZBUF ! type space ;

\ *** 3.4 ACCUMULATOR ARITHMETIC
---
The discussion for b's in Sec. 2.4, ACCUMULATOR ARITHMETIC,
applies to q's as well, except that there is only one
accumulator, QA.
---

CREATE qa /MPQ allot qa mpq_init

: 0qa   ( -- )	       qa mpq_clear qa mpq_init ;
: nu>qa ( num den -- ) 0qa qa -rot mpq_set_si ;
: uu>qa ( num den -- ) 0qa qa -rot mpq_set_ui ;
: df>qa (f: df -- )    0qa qa mpq_set_d ;
: q>qa  (q: q -- )     0qa qa qs@ mpq_set qdrop ;

\ Assume that QA has been initialized.
: qa>q   (q: -- qa )  qa >qs-copy ;
: qa>df   (f: -- df ) qa mpq_get_d ;
: qacanon ( -- )      qa mpq_canonicalize ;

: qa. ( -- ) qa>q q. ;

\ elementary
: -qa	 ( -- )	      qa dup mpq_neg ;
: |qa|	 ( -- )	      qa dup mpq_abs ;
: 1/qa	 ( -- )	      qa dup mpq_inv ;
: qa+	 (q: q -- )   qa dup qs@ mpq_add qdrop ;
: qa-	 (b: b -- )   qa dup qs@ mpq_sub qdrop ;
: qa*	 (b: b -- )   qa dup qs@ mpq_mul qdrop ;
: qa/	 (q: q -- )   qa dup qs@ mpq_div qdrop ;
: qa2^u* ( #bits -- ) >r qa dup r> mpq_mul_2exp ;
: qa2^u/ ( #bits -- ) >r qa dup r> mpq_div_2exp ;

\ *** 3.5 ARITHMETIC

\ private
: nu>'q ( num den 'q -- ) -rot mpq_set_si ;
: uu>'q ( num den 'q -- ) -rot mpq_set_ui ;
: df>'q (f: df s: 'q -- ) mpq_set_d ;
: q>q'  (q: q s: 'q -- )  qs@ mpq_set qdrop ;

: nu>q ( num den -- q: n/d ) ['] nu>'q num-xt>qs ;
: uu>q ( num den -- q: n/d ) ['] uu>'q num-xt>qs ;
: df>q (f: df -- q: df )     ['] df>'q num-xt>qs ;
: qcanon (q: q -- q' )	     ['] q>q'  num-xt>qs ;

: q>df (q: q -- f: df ) qs@ mpq_get_d qdrop ;

\ elementary
: -q	(q: q -- -q )			 q>qa -qa qa>q ;
: |q|	(q: q -- |q|)			 q>qa |qa| qa>q ;
: 1/q	(q: -- 1/q )	      		 q>qa 1/qa qa>q ;
: q+	(q: q1 q2 -- q1+q2 )		 q>qa qa+ qa>q ;
: q-	(q: q1 q2 -- q1-q2 )		 q>qa qa- -qa qa>q ;
: q*	(q: q1 q2 -- q1*q2 )		 q>qa qa* qa>q ;
: q/	(q: q1 q2 -- q1/q2 )		 qswap q>qa qa/ qa>q ;
: q2^u* (q: q s: #bits -- q: q*2^#bits ) q>qa qa2^u* qa>q ;
: q2^u/ (q: q s: #bits -- q: q/2^#bits ) q>qa qa2^u/ qa>q ;

\ *** 3.6 COMPARISONS

: qsgn (q: q -- s: sgn[q] ) qs@ mpq_sgn qdrop ;

: q>  (q: q1 q2 -- s: [q1>q2] )  qs2@ mpq_cmp 0>  q2drop ;
: q<  (q: q1 q2 -- s: [q1<q2] )  qs2@ mpq_cmp 0<  q2drop ;
: q=  (q: q1 q2 -- s: [q1=q2] )  qs2@ mpq_cmp 0=  q2drop ;
: q<= (q: q1 q2 -- s: [q1<=q2] ) qs2@ mpq_cmp 0<= q2drop ;
: q>= (q: q1 q2 -- s: [q1>=q2] ) qs2@ mpq_cmp 0>= q2drop ;

: q0>  (q: q -- s: [q>0] )  qs@ 0 1 mpq_cmp_si 0> qdrop ;
: q0=  (q: q -- s: [q=0] )  qs@ 0 1 mpq_cmp_si 0= qdrop ;
: q0<  (q: q -- s: [q=0] )  qs@ 0 1 mpq_cmp_si 0< qdrop ;
: q0>= (q: q -- s: [q>=0] ) qs@ 0 1 mpq_cmp_si dup 0> swap 0= or qdrop ;
: q0<= (q: q -- s: [q<=0] ) qs@ 0 1 mpq_cmp_si dup 0< swap 0= or qdrop ;

: qnu>  (q: q s: sn ud -- [q>sn/ud] )  qs@ -rot mpq_cmp_si 0> qdrop ;
: qnu=  (q: q s: sn ud -- [q=sn/ud] )  qs@ -rot mpq_cmp_si 0= qdrop ;
: qnu<  (q: q s: sn ud -- [q<sn/ud] )  qs@ -rot mpq_cmp_si 0< qdrop ;
: qnu>= (q: q s: sn ud -- [q>=sn/ud] ) qs@ -rot mpq_cmp_si dup 0> swap 0= or qdrop ;
: qnu<= (q: q s: sn ud -- [q<=sn/ud] ) qs@ -rot mpq_cmp_si dup 0< swap 0= or qdrop ;

: quu>  (q: q s: un ud -- [q>un/ud] )  qs@ -rot mpq_cmp_ui 0> qdrop ;
: quu=  (q: q s: un ud -- [q=un/ud] )  qs@ -rot mpq_cmp_ui 0= qdrop ;
: quu<  (q: q s: un ud -- [q<un/ud] )  qs@ -rot mpq_cmp_ui 0< qdrop ;
: quu>= (q: q s: un ud -- [q>=un/ud] ) qs@ -rot mpq_cmp_ui dup 0> swap 0= or qdrop ;
: quu<= (q: q s: un ud -- [q<=un/ud] ) qs@ -rot mpq_cmp_ui dup 0< swap 0= or qdrop ;


\ *** 4 FLOATING POINT

\ *** 4.1 DEFAULT PRECISION
---
The default precision must be set before any operation that
initializes an mfrec without copying to it from an already
initialized mfrec.  It gets set at the end of this section.

The precision of a dynamic or constant mf never changes after
being set initially, while that of the accumulator can be
changed by the user.  Here's the list of words that initialize
records with the default precision:

  0fa  n>fa  u>fa  df>fa  usqrt>fa  b>fa  q>fa  ba>fa  qa>fa
  mf"  n>mf  u>mf  df>mf  usqrt>mf  b>mf  q>mf  ba>mf  qa>mf

In the current version of GMP, limbs normally use 32 or 64 bits,
and the minimum precision for 32-bit limbs corresponds to 2
limbs, or 64 bits.
---

: set-default-mfprecision ( #bits -- ) mpf_set_default_prec ;
: get-default-mfprecision ( -- #bits ) mpf_get_default_prec ;

128 constant QUAD-MFPRECISION
QUAD-MFPRECISION set-default-mfprecision 

\ *** 4.2 RECORD SPACE

/MPF 2 cells + 4000 * CONSTANT DEFAULT-/MFBUF

: copy-mfrec ( 'f.src 'f.targ -- )
  ---
    Assume that memory for the target mf has been reserved.
    Initialize it with the source precision, and set it to the
    same value.  Intended only for internal use by gcrstack.fs.
  ---
  swap 2dup mpf_get_prec mpf_init2 mpf_set ;

0 value MFSPACE

: fa-start-precision ( -- addr ) mfspace >gextra1 ;

: fa-init-precision ( --  ) get-default-mfprecision fa-start-precision ! ;

: make-mfspace  ( /buf -- )
  ---
    This word writes over the value MFSPACE.  Be sure that it's
    either saved or no longer needed.  It sets the initial
    accumulator precision to the current default.  The space is
    reserved with ALLOCATE, and can be released with MFSPACE
    FREE.  However, that should be done only after executing
    FREE-MFRECS.  See below.
  ---
  ['] mpf_clear ['] copy-mfrec rot /MPF swap
  make-gspace to mfspace
  fa-init-precision ;

DEFAULT-/MFBUF make-mfspace

: free-mfrecs ( -- )
  ---
    Free any unreleased records in mfspace, zero any mfvariables
    pointing to them, and clear mfspace, including the mfstack.
  ---
  mfspace free-grecs ;

\ *** 4.3 GARBAGE COLLECTION

: mfgarbage?	    ( -- flag ) mfspace ggarbage? ;
: collect-mfgarbage ( -- flag ) mfspace collect-ggarbage ;
: mfgc-off	    ( -- )      mfspace ggc-off ;
: mfgc-on	    ( -- )      mfspace ggc-on ;
: mfgc-lock@	    ( -- flag ) mfspace ggc-lock@ ;
: mfgc-lock!	    ( flag -- ) mfspace ggc-lock! ;

\ *** 4.4 STACK

PFE-HOST [IF]
  synonym (mf: (
  synonym MFVARIABLE GVARIABLE
[THEN]
GFORTH-HOST [IF]
  ' (         alias (mf: immediate
  ' GVARIABLE alias MFVARIABLE
[THEN]
IFORTH-HOST [IF]
  : (mf:  postpone ( ; immediate
  : MFVARIABLE  GVARIABLE ;
[THEN]

: mfunused ( -- u )	   mfspace gunused ;
: mfdepth  ( -- u )	   mfspace gdepth ;
: (>mfs)   ( 'f -- mf: f ) mfspace (>gs) ;
: (mfdrop) (mf: f -- )	   mfspace (gdrop) ;

: mfs@	    (mf: f -- f s: fra )		       mfspace gs@ ;
: mfs2@	    (mf: f1 f2 -- f1 f2 s: 'f1 'f2 )	       mfspace gs2@ ;
: mfs3@	    (mf: f1 f2 f3 -- f1 f2 f3 s: 'f1 'f2 'f3 ) mfspace gs3@ ;
: >mfs	    ( 'f -- mf: f )			       mfspace >gs ;
: >mfs-copy ( 'f -- mf: f )			       mfspace >gs-copy ;

: num-xt>mfs  ( <num> xt -- mf: f )
  ---
    Use xt to transfer <num>, a number on the data,
    floating-point, b, or q stack, or an mf number string in the
    input stream, into a dmfrec using the current default
    precision, and push it onto the mfstack.  The stack effect
    for xt is ( <num> 'f -- ).
  ---
  >r 
  mfspace >/grec @ cell+ mfspace ?GROOM
  mfspace new-grec ( 'f) >r
  r@ mpf_init
  2r@ swap EXECUTE
  r@ mfspace (gpush)
  mfspace >gsp @ r> cell- !  \ link mfrec to mfstack
  rdrop ;

\ private
: mf">'f  ( "<chars><">" 'f -- )   \ conversion defends on BASE
  dup postpone mz" ( 'f 'zstr) base @ mpf_set_str
  IF ( 'f) dup mpf_clear
           cell- mfspace >gbrkp !  \ restore mfspace state 
     true ABORT" ***invalid gmp float string" THEN
  ( 'f) drop ;

: mf" ( "<chars><">" -- mf: f ) ['] mf">'f num-xt>mfs ;

0 [IF]  \ simple, unlinked version
: MFCONSTANT ( "name" mf: f -- )
  mfspace gs>extern ( ra) create ,
DOES>  (mf: -- f.ext )  @ (>mfs) ;
[THEN]

0 value LAST-MFCONSTANTP
: MFCONSTANT  ( "name" mf: f -- )
  mfspace gs>extern ( ra) >r create here r> ,
  last-mfconstantp , to last-mfconstantp
DOES>  (mf: -- f.ext )  @ (>mfs) ;

: free-mfconstants  ( -- )
  last-mfconstantp
  BEGIN  dup 
  WHILE  dup @ mpf_clear cell+ @
  REPEAT drop ; 

: mf,'f	  (mf: f -- s: 'f )		      mfspace g,ra ;
: mf@	  ( fva -- mf: f )		      mfspace g@ ;
: mf!	  (mf: f s: mfva -- )		      mfspace g! ;
: mfdrop  (mf: f -- )			      mfspace gdrop ; 
: mf2drop (mf: mf1 f2 -- )		      mfspace g2drop ;
: mf3drop (mf: f1 f2 mf3 -- )		      mfspace g3drop ;
: mfdup   (mf: f -- f f )		      mfspace gdup ;
: mf2dup  (mf: f1 f2 -- f1 f2 f1 f2 )	      mfspace g2dup ;
: mfswap  (mf: f1 f2 -- f2 f1 )		      mfspace gswap ;
: mfnip   (mf: f1 f2 -- f2 )		      mfspace gnip ;
: mfover  (mf: f1 f2 -- f1 f2 f1 )	      mfspace gover ;
: mftuck  (mf: f1 f2 -- f2 f1 f2 )	      mfspace gtuck ;
: mfpick  ( u mf: fu ... f0 -- fu ... f0 fu ) mfspace gpick ;

VARIABLE MFEXP

: >MFSTR  ( u 'f -- )
  over cell+ 2 chars + ?MZBUF
  2>r MZBUF cell+ MFEXP base @ 2r> mpf_get_str zstrlen MZBUF ! ;

: MFSTR.  ( -- )
  BASE @ >r
  MZBUF mcount
  over c@ [char] - =
  IF ." -0." 1 cut-first ELSE ." 0." THEN
  type
  r@ CASE
    10 OF [char] E ENDOF
    16 OF [char] p decimal ENDOF
    [char] @ swap
  ENDCASE
  emit MFEXP @ .
  r> BASE ! ;

: mf. (mf: f s: u -- ) mfs@ >MFSTR MFSTR. mfdrop ;

\ *** 4.5 ACCUMULATOR ARITHMETIC
---
The discussion for b's in Sec. 2.4, ACCUMULATOR ARITHMETIC,
applies to mf's as well, except that there is only one
accumulator, FA.
---

CREATE fa /MPF allot fa mpf_init

: fa-free ( -- ) fa fa-start-precision @ mpf_set_prec_raw fa mpf_clear ;  

: 0fa   ( -- )	    fa-free fa mpf_init fa-init-precision ;
: n>fa  ( n -- )    fa-free fa swap mpf_init_set_si fa-init-precision ;
: u>fa  ( u -- )    fa-free fa swap mpf_init_set_ui fa-init-precision ;
: df>fa (f: df -- ) fa-free fa mpf_init_set_d  fa-init-precision ;

: mf>fa  (mf: f -- )
  fa-free fa mfs@ mpf_init_set mfdrop
  fa mpf_get_prec fa-start-precision ! ;

\ Assume that FA has been initialized.

: fa>mf	    (mf: -- a )	      fa >mfs-copy ;
: fa>n	    ( -- n )	      fa mpf_get_si ;
: fa>u	    ( -- u )	      fa mpf_get_ui ;
: fa>df	    (f: -- df )	      fa mpf_get_d ;
: fa>df-exp ( 'exp -- f: df ) fa mpf_get_d_2exp ;

: fa. ( u -- ) fa >MFSTR MFSTR. ;

\ Support for changing fa precision on the fly.
: set-fa-precision ( #bits -- ) fa swap mpf_set_prec_raw ;
: get-fa-precision ( -- #bits ) fa mpf_get_prec ;

\ elementary
: -fa	   ( -- )	 fa dup mpf_neg ;
: |fa|	   ( -- )	 fa dup mpf_abs ;
: fa+	   (mf: f -- )	 fa dup mfs@ mpf_add mfdrop ;
: fa-	   (mf: f -- )	 fa dup mfs@ mpf_sub mfdrop ;
: fau+	   ( u -- )	 fa dup  rot mpf_add_ui ;
: fau-	   ( u -- )	 fa dup  rot mpf_sub_ui ;
: ufa-	   ( u -- )	 fa dup -rot mpf_ui_sub ;
: fa*	   (mf: f -- )	 fa dup mfs@ mpf_mul mfdrop ;
: fau*	   ( u -- )	 fa dup rot mpf_mul_ui ;
: fa/	   (mf: f -- )	 fa dup mfs@ mpf_div mfdrop ;
: fau/	   ( u -- )	 fa dup rot mpf_div_ui ;
: ufa/	   ( u -- )	 fa swap fa mpf_ui_div ;
: fasqrt   ( -- )	 fa dup mpf_sqrt ;
: usqrt>fa ( u -- )	 0fa fa swap mpf_sqrt_ui ;
: fa^u	   ( u -- )	 fa dup rot mpf_pow_ui ;
: fa2^u*   ( u -- )	 >r fa dup r> mpf_mul_2exp ;
: fa/2^u   ( u -- )	 >r fa dup r> mpf_div_2exp ;
: farel-   (mf: ref -- ) fa dup mfs@ swap mpf_reldiff mfdrop ; 

\ integer rounding
: fa-ceil    ( -- )	 fa dup mpf_ceil ;
: fa-floor   ( -- )	 fa dup mpf_floor ;
: fa-trunc   ( -- )	 fa dup mpf_trunc ;
: fa-integer ( -- flag ) fa mpf_integer_p wff ;

\ *** 4.6 ARITHMETIC

\ private
: n>'f     ( n 'f -- )	     swap mpf_set_si ;
: u>'f     ( u 'f -- )	     swap mpf_set_ui ;
: df>'f    (f: df s: 'f -- ) mpf_set_d ;
: usqrt>'f ( u 'f -- )	     swap mpf_sqrt_ui ;

: n>mf  ( n -- mf: n )	   ['] n>'f  num-xt>mfs ;
: u>mf  ( u -- mf: u )	   ['] u>'f  num-xt>mfs ;
: df>mf (f: df -- mf: df ) ['] df>'f num-xt>mfs ;

: mf>n	    (mf: f -- s: n )		mfs@ mpf_get_si mfdrop ; 
: mf>u	    (mf: f -- s: |f| )		mfs@ mpf_get_ui mfdrop ;
: mf>df	    (mf: f -- f: f )		mfs@ mpf_get_d mfdrop ;
: mf>df-exp (mf: f s: 'exp -- f: frac ) mfs@ mpf_get_d_2exp mfdrop ;

\ elementary

: -mf  (mf: f -- -f )		mf>fa -fa  fa>mf ;
: |mf| (mf: f -- |f|)		mf>fa |fa| fa>mf ;
: mf+  (mf: f1 f2 -- f1+f2 )	mf>fa fa+  fa>mf ;
: mf-  (mf: f1 f2 -- f1-f2 )	mf>fa fa- -fa fa>mf ;
: mfu+ (mf: f s: u -- mf: f+u )	mf>fa fau+ fa>mf ;
: mfu- (mf: f s: u -- mf: f-u )	mf>fa fau- fa>mf ;
: umf- (mf: f s: u -- mf: u-f )	mf>fa ufa- fa>mf ;
: mf*  (mf: f1 f2 -- f1*f2 )	mf>fa fa*  fa>mf ;
: mfu* (mf: f s: u -- mf: f*u )	mf>fa fau* fa>mf ;

: mf/	   (mf: f1 f2 -- f1*f2 )    mfswap mf>fa fa/ fa>mf ;
: mfu/	   (mf: f s: u -- mf: f/u ) mf>fa fau/ fa>mf ;
: umf/	   ( u mf: f -- u/f )	    mf>fa ufa/ fa>mf ;
: mfsqrt   (mf: f -- sqrt[f] )	    mf>fa fasqrt fa>mf ;
: usqrt>mf ( u -- mf: sqrt[u] )	    ['] usqrt>'f num-xt>mfs ;

: mf^u	 ( u mf: f -- f^u )	    mf>fa fa^u fa>mf ;
: mf*2^u (mf: f s: u -- mf: f*2^u ) mf>fa fa2^u* fa>mf ;
: mf/2^u (mf: f s: u -- mf: f*2^u ) mf>fa fa/2^u fa>mf ;
: mfrel- (mf: meas ref -- reldif )  mfswap mf>fa farel- fa>mf ;

: mf-ceil    (mf: f -- ceil[f] )  mf>fa fa-ceil fa>mf ;
: mf-floor   (mf: f -- floor[f] ) mf>fa fa-floor fa>mf ;
: mf-trunc   (mf: f -- trunc[f] ) mf>fa fa-trunc fa>mf ;
: mf-integer (mf: f -- s: flag )  mfs@ mpf_integer_p wff mfdrop ;

\ *** 4.7 COMPARISONS

: mfsgn (mf: f -- s: sgn[f] ) mfs@ mpf_sgn mfdrop ;

: mf>  (mf: f1 f2 -- s: [f1>f2] )  mfs2@ mpf_cmp 0>  mf2drop ;
: mf<  (mf: f1 f2 -- s: [f1<f2] )  mfs2@ mpf_cmp 0<  mf2drop ;
: mf=  (mf: f1 f2 -- s: [f1=f2] )  mfs2@ mpf_cmp 0=  mf2drop ;
: mf<= (mf: f1 f2 -- s: [f1<=f2] ) mfs2@ mpf_cmp 0<= mf2drop ;
: mf>= (mf: f1 f2 -- s: [f1>=f2] ) mfs2@ mpf_cmp 0>= mf2drop ;

: mf0>  (mf: f -- s: [f>0] )  mfs@ 0 mpf_cmp_si 0> mfdrop ;
: mf0=  (mf: f -- s: [f=0] )  mfs@ 0 mpf_cmp_si 0= mfdrop ;
: mf0<  (mf: f -- s: [f=0] )  mfs@ 0 mpf_cmp_si 0< mfdrop ;
: mf0>= (mf: f -- s: [f>=0] ) mfs@ 0 mpf_cmp_si dup 0> swap 0= or mfdrop ;
: mf0<= (mf: f -- s: [f<=0] ) mfs@ 0 mpf_cmp_si dup 0< swap 0= or mfdrop ;

: mfn>  (mf: f s: n -- [f>n] )  mfs@ swap mpf_cmp_si 0> mfdrop ;
: mfn=  (mf: f s: n -- [f=n] )  mfs@ swap mpf_cmp_si 0= mfdrop ;
: mfn<  (mf: f s: n -- [f<n] )  mfs@ swap mpf_cmp_si 0< mfdrop ;
: mfn>= (mf: f s: n -- [f>=n] ) mfs@ swap mpf_cmp_si dup 0> swap 0= or mfdrop ;
: mfn<= (mf: f s: n -- [f<=n] ) mfs@ swap mpf_cmp_si dup 0< swap 0= or mfdrop ;

: mfu>  (mf: f s: u -- [f>u] )  mfs@ swap mpf_cmp_ui 0> mfdrop ;
: mfu=  (mf: f s: u -- [f=u] )  mfs@ swap mpf_cmp_ui 0= mfdrop ;
: mfu<  (mf: f s: u -- [f<u] )  mfs@ swap mpf_cmp_ui 0< mfdrop ;
: mfu>= (mf: f s: u -- [f>=u] ) mfs@ swap mpf_cmp_ui dup 0> swap 0= or mfdrop ;
: mfu<= (mf: f s: u -- [f<=u] ) mfs@ swap mpf_cmp_ui dup 0< swap 0= or mfdrop ;

: mfdf>  (mf: f f: df -- s: [f>df] )  mfs@ mpf_cmp_d 0> mfdrop ;
: mfdf=  (mf: f f: df -- s: [f=df] )  mfs@ mpf_cmp_d 0= mfdrop ;
: mfdf<  (mf: f f: df -- s: [f<df] )  mfs@ mpf_cmp_d 0< mfdrop ;
: mfdf>= (mf: f f: df -- s: [f>=df] ) mfs@ mpf_cmp_d dup 0> swap 0= or mfdrop ;
: mfdf<= (mf: f f: df -- s: [f<=df] ) mfs@ mpf_cmp_d dup 0< swap 0= or mfdrop ;


\ *** 5 MIXED STACK WORDS

\ private
: q>'b  (q: q s: 'b -- ) qs@  mpz_set_q qdrop ;
: f>'b (mf: f s: 'b -- ) mfs@ mpz_set_f mfdrop ;
: b>'q (b: b s: 'q -- )  bs@  mpq_set_z bdrop ;
: f>'q (f: f s: 'q -- )  mfs@ mpq_set_f mfdrop ; 
: b>'f  (b: b s: 'f -- ) bs@  mpf_set_z bdrop ;
: q>'f  (q: q s: 'f -- ) qs@  mpf_set_q qdrop ;

: q>b  (q: q -- b: [q] )  ['] q>'b num-xt>bs ;
: mf>b (mf: f -- b: [f] ) ['] f>'b num-xt>bs ;
: b>q  (b: b -- q: b )    ['] b>'q num-xt>qs ;
: mf>q (mf: f -- q: f )   ['] f>'q num-xt>qs ;
: b>mf (b: b -- mf: b )   ['] b>'f num-xt>mfs ;
: q>mf (q: q -- mf: q )   ['] q>'f num-xt>mfs ;

\ private
: qa>'b ( 'b -- ) qa mpz_set_q ;
: fa>'b ( 'b -- ) fa mpz_set_f ;
: ba>'q ( 'q -- ) ba mpq_set_z ;
: fa>'q ( 'q -- ) fa mpq_set_f ;
: ba>'f ( 'f -- ) ba mpf_set_z ;
: qa>'f ( 'f -- ) qa mpf_set_q ;

: qa>b  (b: -- [qa] ) ['] qa>'b num-xt>bs ;
: fa>b  (b: -- [fa] ) ['] fa>'b num-xt>bs ;
: ba>q  (q: -- ba )   ['] ba>'q num-xt>qs ;
: fa>q  (q: -- fa )   ['] fa>'q num-xt>qs ;
: ba>mf (mf: -- ba )  ['] ba>'f num-xt>mfs ;
: qa>mf (mf: -- qa )  ['] qa>'f num-xt>mfs ;

: q>ba  (q: q -- )  0ba ba qs@  mpz_set_q qdrop ;
: mf>ba (mf: f -- ) 0ba ba mfs@ mpz_set_f mfdrop ;
: b>qa  (b: b -- )  0qa qa bs@  mpq_set_z bdrop ;
: mf>qa (mf: f -- ) 0qa qa mfs@ mpq_set_f mfdrop ;
: b>fa  (b: b -- )  0fa fa bs@  mpf_set_z bdrop ;
: q>fa  (q: q -- )  0fa fa qs@  mpf_set_q qdrop ;

: qa>ba (q: q -- )  0ba ba qa mpz_set_q ;
: fa>ba (mf: f -- ) 0ba ba fa mpz_set_f ;
: ba>qa (b: b -- )  0qa qa ba mpq_set_z ;
: fa>qa (mf: f -- ) 0qa qa fa mpq_set_f ;
: ba>fa (b: b -- )  0fa fa ba mpf_set_z ;
: qa>fa (q: q -- )  0fa fa qa mpf_set_q ;

\ private
: qnum>'b  (q: q s: 'b -- ) qs@ mpq_get_num qdrop ;
: qden>'b  (q: q s: 'b -- ) qs@ mpq_get_den qdrop ;
: qanum>'b ( 'b -- )	    qa  mpq_get_num ;
: qaden>'b ( 'b -- )	    qa  mpq_get_den ;
: b>'qn (b: num s: q' -- )  dup qs@ mpq_set qdrop bs@ mpq_set_num bdrop ;
: b>'qd (b: num s: q' -- )  dup qs@ mpq_set qdrop bs@ mpq_set_den bdrop ;
: ba>'qn (q: q s: 'q' -- )  dup qs@ mpq_set qdrop ba mpq_set_num ; 
: ba>'qd (q: q s: 'q' -- )  dup qs@ mpq_set qdrop ba mpq_set_den ; 

: qnum>b (q: n/d -- b: n )	 ['] qnum>'b num-xt>bs ;
: qden>b (q: n/d -- b: d )	 ['] qden>'b num-xt>bs ;
: qanum>b ( -- num )		 ['] qanum>'b num-xt>bs ;
: qaden>b ( -- den )		 ['] qaden>'b num-xt>bs ;
: b>qnum (b: n' q: n/d -- n'/d ) ['] b>'qn num-xt>qs ;
: b>qden (b: d' q: n/d -- n/d' ) ['] b>'qd num-xt>qs ;
: ba>qnum (q: -- n'/d )		 ['] ba>'qn num-xt>qs ;
: ba>qden (q: -- n/d' )		 ['] ba>'qd num-xt>qs ;

: qnum>ba (q: n/d -- ) ba qs@ mpq_numref mpz_init_set qdrop ;
: qden>ba (q: n/d -- ) ba qs@ mpq_denref mpz_init_set qdrop ;
: qanum>ba ( -- )      ba qa  mpq_numref mpz_init_set ;
: qaden>ba ( -- )      ba qa  mpq_denref mpz_init_set ;
: b>qanum (b: n -- )   qa bs@ mpq_set_num bdrop ;
: b>qaden (b: d -- )   qa bs@ mpq_set_den bdrop ;
: ba>qanum  ( -- )     qa ba mpq_set_num ;
: ba>qaden  ( -- )     qa ba mpq_set_den ;


\ *** 6 RANDOM NUMBERS

\ *** 6.1 STATE
---
State words use only the data stack.
---

0 value last-randstate

: RANDSTATE:  \ compile: ( "name" -- )
              \ run:     ( -- 'rand )
  ---
    Allot an unitialized randstate and link it into the list of
    randstates.
  ---
\ create /GMP-RANDSTATE aligned allot last-randstate , ;
  create here /GMP-RANDSTATE aligned allot last-randstate ,
  ( here) to last-randstate ;

: free-randstate ( 'rand -- ) gmp_randclear ;

: free-randstates  ( -- )
  ---
    Free the randstates in the RANDVAR list.  Do not use this
    word unless all RANDVAR's are initialized.
  ---
  last-randstate
  BEGIN  dup 
  WHILE  dup gmp_randclear /GMP-RANDSTATE aligned + @
  REPEAT to last-randstate ;

: copy-rand ( 'rand.src 'rand.targ -- ) swap gmp_randinit_set ;

\ *** 6.2 INITIALIZATION
---
All words use only the data stack except for RANDINIT-LC-2^U,
which also uses the bstack.
---

\ Mersenne twister in gmp 5.0.1
: randinit-default ( 'rand -- )	gmp_randinit_default ;

\ Mersenne twister
: randinit-mt ( 'rand -- ) gmp_randinit_mt ;

\ Linear congruential:  X = (aX + c) mod 2^u
: randinit-lc-2^u (b: a s: c u 'rand -- )
  -rot 2>r bs@ 2r> ( 'rand 'b c u) gmp_randinit_lc_2exp bdrop ;

\ Linear congruential:  table driven a, c, u > 2*#bits
\			u <= 128 in gmp 5.0.1
: randinit-lc-bits ( min.#bits 'rand -- flag ) swap gmp_randinit_lc_2exp_size wff ;

\ *** 6.3 SEEDING
---
Seeds are set from the data stack or the bstack.
---

: b>seed (b: seed s: 'rand -- ) bs@ gmp_randseed bdrop ;
: u>seed ( u 'rand -- )		swap gmp_randseed_ui ;

\ *** 6.4 GENERATION
---
Random numbers are generated on the data stack, the bstack, or
the mfstack, or in the b or mf accumulator.
---

\ private
: rand>'b  ( 'rand u 'b -- )	 -rot mpz_urandomb ;
: rand2>'b ( 'rand u 'b -- )	 -rot mpz_rrandomb ;
: rand>'f  ( 'rand sbits 'f -- ) -rot mpf_urandomb ;
: rand2>'f ( limbs p 'f -- )	 -rot mpf_random2 ;

\ 0 <= rand <= 2^u - 1, 0 <= u <= bits/cell
: rand-#bits>u   ( u 'rand -- rand )	swap gmp_urandomb_ui ;
: rand-#bits>ba  ( u 'rand --  )	swap ba -rot mpz_urandomb ;
: rand-#bits>b	 ( u 'rand -- b: rand ) swap ['] rand>'b num-xt>bs ;
: rand2-#bits>ba ( u 'rand --  )	swap ba -rot mpz_rrandomb ;
: rand2-#bits>b  ( u 'rand -- b: rand ) swap ['] rand2>'b num-xt>bs ;

\ 0 <= rand < 1,  sbits significant bits
: rand-sbits>fa ( sbits 'rand -- )	    swap fa -rot mpf_urandomb ;
: rand-sbits>mf	( sbits 'rand -- fs: rand ) swap ['] rand>'f num-xt>mfs ;

\ -1.0 x lbase^-p < |rand| < 1.0 x lbase^p [inequality?!]
\	sgn[rand] = sgn[max#limbs]
\       note the absence of 'rand
: rand2-limbs>fa ( max#limbs +p -- )	      fa -rot mpf_random2 ;
: rand2-limbs>mf ( max#limbs +p -- fs: rand ) ['] rand2>'f num-xt>mfs ;

\ 0<= rand < n,  rand and n are u's
: rand[0,n-1]>u ( +n 'rand -- rand ) swap gmp_urandomm_ui ;
