/**
 *       Title:  GNU Multiple Precision PFE module
 *        File:  libgmp-pfe.h
 *    Log file:  gmpfr.log
 *  Test files:  libgmp-test.fs, gmpfr-test.fs
 *      Author:  David N. Williams
 *     License:  LGPL
 *     Version:  0.8.0
 *     Started:  January 16, 2011
 *     Revised:  March 26, 2011
 *
 * Copyright (C) 2011 by David N. Williams
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later
 * version.
 *
 * This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU Library General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA.
 *
 * If you take advantage of the option in the LGPL to put a
 * particular version of this library under the GPL, the author
 * would regard it as polite if you would put any direct
 * modifications under the LGPL as well, and include a copy of
 * this request near the beginning of the modified library
 * source.  A "direct modification" is one that enhances or
 * extends the library in line with its original concept, as
 * opposed to developing a distinct application or library which
 * might use it.
 */

#ifndef _LIBGMP_PFE_H 
#define _LIBGMP_PFE_H

// abbreviations
#define slong    signed long
#define ulong    unsigned long

typedef __gmp_randstate_struct* gmp_rstate_t;

/***********************
   Functions and Words
 ************************/

// 1 General support
FCode (p4_gmp_version);
FCode (p4_mp_bits_per_limb);
// 2 Integers
// 2.1 Initialization
FCode (p4_mpz_init);
FCode (p4_mpz_clear);
// 2.2 Assignment
FCode (p4_mpz_set);
FCode (p4_mpz_set_ui);
FCode (p4_mpz_set_si);
FCode (p4_mpz_set_d);
FCode (p4_mpz_set_q);
FCode (p4_mpz_set_f);
FCode (p4_mpz_set_str);
FCode (p4_mpz_swap);
// 2.3 Combined Initialization and Assignment
FCode (p4_mpz_init_set);
FCode (p4_mpz_init_set_ui);
FCode (p4_mpz_init_set_si);
FCode (p4_mpz_init_set_d);
FCode (p4_mpz_init_set_str);
// 2.4 Conversion
FCode (p4_mpz_get_ui);
FCode (p4_mpz_get_si);
FCode (p4_mpz_get_d);
FCode (p4_mpz_get_d_2exp);
FCode (p4_mpz_get_str);
// 2.5 Arithmemtic
FCode (p4_mpz_add);
FCode (p4_mpz_add_ui);
FCode (p4_mpz_sub);
FCode (p4_mpz_sub_ui);
FCode (p4_mpz_ui_sub);
FCode (p4_mpz_mul);
FCode (p4_mpz_mul_si);
FCode (p4_mpz_mul_ui);
FCode (p4_mpz_addmul);
FCode (p4_mpz_addmul_ui);
FCode (p4_mpz_submul);
FCode (p4_mpz_submul_ui);
FCode (p4_mpz_mul_2exp);
FCode (p4_mpz_neg);
FCode (p4_mpz_abs);
// 2.6 Division
FCode (p4_mpz_cdiv_q);
FCode (p4_mpz_cdiv_r);
FCode (p4_mpz_cdiv_qr);
FCode (p4_mpz_cdiv_q_ui);
FCode (p4_mpz_cdiv_r_ui);
FCode (p4_mpz_cdiv_qr_ui);
FCode (p4_mpz_cdiv_ui);
FCode (p4_mpz_cdiv_q_2exp);
FCode (p4_mpz_cdiv_r_2exp);
FCode (p4_mpz_fdiv_q);
FCode (p4_mpz_fdiv_r);
FCode (p4_mpz_fdiv_qr);
FCode (p4_mpz_fdiv_q_ui);
FCode (p4_mpz_fdiv_r_ui);
FCode (p4_mpz_fdiv_qr_ui);
FCode (p4_mpz_fdiv_ui);
FCode (p4_mpz_fdiv_q_2exp);
FCode (p4_mpz_fdiv_r_2exp);
FCode (p4_mpz_tdiv_q);
FCode (p4_mpz_tdiv_r);
FCode (p4_mpz_tdiv_qr);
FCode (p4_mpz_tdiv_q_ui);
FCode (p4_mpz_tdiv_r_ui);
FCode (p4_mpz_tdiv_qr_ui);
FCode (p4_mpz_tdiv_ui);
FCode (p4_mpz_tdiv_q_2exp);
FCode (p4_mpz_tdiv_r_2exp);
FCode (p4_mpz_mod);
FCode (p4_mpz_mod_ui);
FCode (p4_mpz_divexact);
FCode (p4_mpz_divexact_ui);
FCode (p4_mpz_divisible_p);
FCode (p4_mpz_divisible_ui_p);
FCode (p4_mpz_divisible_2exp_p);
FCode (p4_mpz_congruent_p);
FCode (p4_mpz_congruent_ui_p);
FCode (p4_mpz_congruent_2exp_p);
// 2.7  Exponentiation
FCode (p4_mpz_powm);
FCode (p4_mpz_powm_ui);
FCode (p4_mpz_powm_sec);
FCode (p4_mpz_pow_ui);
FCode (p4_mpz_ui_pow_ui);
// 2.8  Root extraction
FCode (p4_mpz_root);
FCode (p4_mpz_rootrem);
FCode (p4_mpz_sqrt);
FCode (p4_mpz_sqrtrem);
FCode (p4_mpz_perfect_power_p);
FCode (p4_mpz_perfect_square_p);
// 2.9  Number theoretics
FCode (p4_mpz_probab_prime_p);
FCode (p4_mpz_nextprime);
FCode (p4_mpz_gcd);
FCode (p4_mpz_gcd_ui);
FCode (p4_mpz_gcdext);
FCode (p4_mpz_lcm);
FCode (p4_mpz_lcm_ui);
FCode (p4_mpz_invert);
FCode (p4_mpz_jacobi);
FCode (p4_mpz_legendre);
FCode (p4_mpz_kronecker);
FCode (p4_mpz_kronecker_si);
FCode (p4_mpz_kronecker_ui);
FCode (p4_mpz_si_kronecker);
FCode (p4_mpz_ui_kronecker);
FCode (p4_mpz_remove);
FCode (p4_mpz_fac_ui);
FCode (p4_mpz_bin_ui);
FCode (p4_mpz_bin_uiui);
FCode (p4_mpz_fib_ui);
FCode (p4_mpz_fib2_ui);
FCode (p4_mpz_lucnum_ui);
FCode (p4_mpz_lucnum2_ui);
// 2.10 Comparison
FCode (p4_mpz_cmp);
FCode (p4_mpz_cmp_d);
FCode (p4_mpz_cmp_si);
FCode (p4_mpz_cmp_ui);
FCode (p4_mpz_cmpabs);
FCode (p4_mpz_cmpabs_d);
FCode (p4_mpz_cmpabs_ui);
FCode (p4_mpz_sgn);
// 2.11 Logic and bit manipulation
FCode (p4_mpz_and);
FCode (p4_mpz_ior);
FCode (p4_mpz_xor);
FCode (p4_mpz_com);
FCode (p4_mpz_popcount);
FCode (p4_mpz_hamdist);
FCode (p4_mpz_scan0);
FCode (p4_mpz_scan1);
FCode (p4_mpz_setbit);
FCode (p4_mpz_clrbit);
FCode (p4_mpz_combit);
FCode (p4_mpz_tstbit);
// 2.12 Input and output
FCode (p4_mpz_out_str);
FCode (p4_mpz_inp_str);
FCode (p4_mpz_out_raw);
FCode (p4_mpz_inp_raw);
// 2.13 Random numbers
FCode (p4_mpz_urandomb);
FCode (p4_mpz_urandomm);
FCode (p4_mpz_rrandomb);
// 2.14 Integer export and import
FCode (p4_mpz_import);
FCode (p4_mpz_export);
// 2.15 Miscellaneous
FCode (p4_mpz_fits_ulong_p);
FCode (p4_mpz_fits_slong_p);
FCode (p4_mpz_fits_uint_p);
FCode (p4_mpz_fits_sint_p);
FCode (p4_mpz_fits_ushort_p);
FCode (p4_mpz_fits_sshort_p);
FCode (p4_mpz_odd_p);
FCode (p4_mpz_even_p);
FCode (p4_mpz_sizeinbase);
// 2.16 Special
FCode (p4_mpz_getlimbn);
FCode (p4_mpz_size);
// 3 Rationals
FCode (p4_mpq_canonicalize);
// 3.1 Initialization and assignment
FCode (p4_mpq_init);
FCode (p4_mpq_clear);
FCode (p4_mpq_set);
FCode (p4_mpq_set_z);
FCode (p4_mpq_set_ui);
FCode (p4_mpq_set_si);
FCode (p4_mpq_set_str);
FCode (p4_mpq_swap);
// 3.2 Conversion
FCode (p4_mpq_get_d);
FCode (p4_mpq_set_d);
FCode (p4_mpq_set_f);
FCode (p4_mpq_get_str);
// 3.3 Arithmetic
FCode (p4_mpq_add);
FCode (p4_mpq_sub);
FCode (p4_mpq_mul);
FCode (p4_mpq_mul_2exp);
FCode (p4_mpq_div);
FCode (p4_mpq_div_2exp);
FCode (p4_mpq_neg);
FCode (p4_mpq_abs);
FCode (p4_mpq_inv);
// 3.4 Comparison
FCode (p4_mpq_cmp);
FCode (p4_mpq_cmp_ui);
FCode (p4_mpq_cmp_si);
FCode (p4_mpq_sgn);
FCode (p4_mpq_equal);
// 3.5 Integer functions
FCode (p4_mpq_numref);
FCode (p4_mpq_denref);
FCode (p4_mpq_get_num);
FCode (p4_mpq_get_den);
FCode (p4_mpq_set_num);
FCode (p4_mpq_set_den);
// 3.6 Input and output
FCode (p4_mpq_out_str);
FCode (p4_mpq_inp_str);
// 4 Floating point
// 4.1 Initialization
FCode (p4_mpf_set_default_prec);
FCode (p4_mpf_get_default_prec);
FCode (p4_mpf_init);
FCode (p4_mpf_init2);
FCode (p4_mpf_clear);
FCode (p4_mpf_get_prec);
FCode (p4_mpf_set_prec);
FCode (p4_mpf_set_prec_raw);
// 4.2 Assignment
FCode (p4_mpf_set);
FCode (p4_mpf_set_ui);
FCode (p4_mpf_set_si);
FCode (p4_mpf_set_d);
FCode (p4_mpf_set_z);
FCode (p4_mpf_set_q);
FCode (p4_mpf_set_str);
FCode (p4_mpf_swap);
// 4.3 Combined initialization and assignment
FCode (p4_mpf_init_set);
FCode (p4_mpf_init_set_ui);
FCode (p4_mpf_init_set_si);
FCode (p4_mpf_init_set_d);
FCode (p4_mpf_init_set_str);
// 4.4 Conversion
FCode (p4_mpf_get_d);
FCode (p4_mpf_get_d_2exp);
FCode (p4_mpf_get_si);
FCode (p4_mpf_get_ui);
FCode (p4_mpf_get_str);
// 4.5 Arithmetic
FCode (p4_mpf_add);
FCode (p4_mpf_add_ui);
FCode (p4_mpf_sub);
FCode (p4_mpf_ui_sub);
FCode (p4_mpf_sub_ui);
FCode (p4_mpf_mul);
FCode (p4_mpf_mul_ui);
FCode (p4_mpf_div);
FCode (p4_mpf_ui_div);
FCode (p4_mpf_div_ui);
FCode (p4_mpf_sqrt);
FCode (p4_mpf_sqrt_ui);
FCode (p4_mpf_pow_ui);
FCode (p4_mpf_neg);
FCode (p4_mpf_abs);
FCode (p4_mpf_mul_2exp);
FCode (p4_mpf_div_2exp);
// 4.6 Comparison
FCode (p4_mpf_cmp);
FCode (p4_mpf_cmp_d);
FCode (p4_mpf_cmp_ui);
FCode (p4_mpf_cmp_si);
FCode (p4_mpf_eq);
FCode (p4_mpf_reldiff);
FCode (p4_mpf_sgn);
// 4.7 Input and output
FCode (p4_mpf_out_str);
FCode (p4_mpf_inp_str);
// 4.8 Miscellaneous
FCode (p4_mpf_ceil);
FCode (p4_mpf_floor);
FCode (p4_mpf_trunc);
FCode (p4_mpf_integer_p);
FCode (p4_mpf_fits_ulong_p);
FCode (p4_mpf_fits_slong_p);
FCode (p4_mpf_fits_uint_p);
FCode (p4_mpf_fits_sint_p);
FCode (p4_mpf_fits_ushort_p);
FCode (p4_mpf_fits_sshort_p);
FCode (p4_mpf_urandomb);
FCode (p4_mpf_random2);
// 5 Low level
// 6 Random number state
// 6.1 Initialization
FCode (p4_gmp_randinit_default);
FCode (p4_gmp_randinit_mt);
FCode (p4_gmp_randinit_lc_2exp);
FCode (p4_gmp_randinit_lc_2exp_size);
FCode (p4_gmp_randinit_set);
FCode (p4_gmp_randclear);
// 6.2 Seeding
FCode (p4_gmp_randseed);
FCode (p4_gmp_randseed_ui);
// 6.3 Miscellaneous
FCode (p4_gmp_urandomb_ui);
FCode (p4_gmp_urandomm_ui);
// 7 Formatted output
// 8 Formatted input

#endif	/* ndef _LIBGMP_PFE_H */
